import os
import shutil
import tempfile
import unittest
import warnings

import numpy as np

import rubin_sim.maf.db as db


class TestResultsDb(unittest.TestCase):
    def setUp(self):
        self.out_dir = "Out"
        self.metric_name = "Count ExpMJD"
        self.slicer_name = "OneDSlicer"
        self.run_name = "fakeopsim"
        self.sql_constraint = ""
        self.info_label = "Dithered"
        self.metric_data_file = "testmetricdatafile.npz"
        self.plot_type = "BinnedData"
        self.plot_name = "testmetricplot_BinnedData.png"
        self.summary_stat_name1 = "Mean"
        self.summary_stat_value1 = 20
        self.summary_stat_name2 = "Median"
        self.summary_stat_value2 = 18
        self.summary_stat_name3 = "TableFrac"
        self.summary_stat_value3 = np.empty(10, dtype=[("name", "|S12"), ("value", float)])
        for i in range(10):
            self.summary_stat_value3["name"] = "test%d" % (i)
            self.summary_stat_value3["value"] = i
        self.display_dict = {
            "group": "seeing",
            "subgroup": "all",
            "order": 1,
            "caption": "lalalalal",
        }

    def test_db_creation(self):
        # Test default sqlite file created.
        tempdir = tempfile.mkdtemp(prefix="resDb")
        resultsdb = db.ResultsDb(out_dir=tempdir)
        self.assertTrue(os.path.isfile(os.path.join(tempdir, "resultsDb_sqlite.db")))
        resultsdb.close()
        # Test that get appropriate exception if directory doesn't exist.
        sqlitefilename = os.path.join(self.out_dir + "test", "testDb_sqlite.db")
        self.assertRaises(ValueError, db.ResultsDb, database=sqlitefilename)
        shutil.rmtree(tempdir)

    def test_add_data(self):
        tempdir = tempfile.mkdtemp(prefix="resDb")
        results_db = db.ResultsDb(out_dir=tempdir)
        # Add metric.
        metric_id = results_db.update_metric(
            self.metric_name,
            self.slicer_name,
            self.run_name,
            self.sql_constraint,
            self.info_label,
            self.metric_data_file,
        )
        # Try to re-add metric
        # (should get back same metric id as previous, with no add).
        metric_id2 = results_db.update_metric(
            self.metric_name,
            self.slicer_name,
            self.run_name,
            self.sql_constraint,
            self.info_label,
            self.metric_data_file,
        )
        self.assertEqual(metric_id, metric_id2)
        run1 = results_db.session.query(db.MetricRow).filter_by(metric_id=metric_id).all()
        self.assertEqual(len(run1), 1)
        # Add plot.
        results_db.update_plot(metric_id, self.plot_type, self.plot_name)
        # Add normal summary statistics.
        results_db.update_summary_stat(metric_id, self.summary_stat_name1, self.summary_stat_value1)
        results_db.update_summary_stat(metric_id, self.summary_stat_name2, self.summary_stat_value2)
        # Add something like tableFrac summary statistic.
        results_db.update_summary_stat(metric_id, self.summary_stat_name3, self.summary_stat_value3)
        # Test get warning when try to add a non-conforming summary stat
        # (not 'name' & 'value' cols).
        teststat = np.empty(10, dtype=[("col", "|S12"), ("value", float)])
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            results_db.update_summary_stat(metric_id, "testfail", teststat)
            self.assertIn("not save", str(w[-1].message))
        # Test we get warning when try to add a string (non-conforming)
        # summary stat.
        teststat = "teststring"
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            results_db.update_summary_stat(metric_id, "testfail", teststat)
            self.assertIn("not save", str(w[-1].message))
        shutil.rmtree(tempdir)


class TestUseResultsDb(unittest.TestCase):
    def setUp(self):
        self.out_dir = "Out"
        self.metric_name = "Count ExpMJD"
        self.slicer_name = "OneDSlicer"
        self.run_name = "fakeopsim"
        self.constraint = ""
        self.info_label = "Dithered"
        self.metric_data_file = "testmetricdatafile.npz"
        self.plot_type = "BinnedData"
        self.plot_name = "testmetricplot_BinnedData.png"
        self.summary_stat_name1 = "Mean"
        self.summary_stat_value1 = 20
        self.summary_stat_name2 = "Median"
        self.summary_stat_value2 = 18
        self.tempdir = tempfile.mkdtemp(prefix="resDb")
        self.results_db = db.ResultsDb(self.tempdir)
        self.metric_id = self.results_db.update_metric(
            self.metric_name,
            self.slicer_name,
            self.run_name,
            self.constraint,
            self.info_label,
            self.metric_data_file,
        )
        self.results_db.update_plot(self.metric_id, self.plot_type, self.plot_name)
        self.results_db.update_summary_stat(self.metric_id, self.summary_stat_name1, self.summary_stat_value1)
        self.results_db.update_summary_stat(self.metric_id, self.summary_stat_name2, self.summary_stat_value2)

    def testget_ids(self):
        mids = self.results_db.get_all_metric_ids()
        self.assertEqual(mids[0], self.metric_id)
        mid = self.results_db.get_metric_id(self.metric_name)
        self.assertEqual(mid[0], self.metric_id)
        mid = self.results_db.get_metric_id("notreal")
        self.assertEqual(len(mid), 0)

    def testshow_summary(self):
        self.results_db.get_summary_stats()

    def tearDown(self):
        self.results_db.close()
        shutil.rmtree(self.tempdir)


if __name__ == "__main__":
    unittest.main()



__all__ = ("PeriodicStar", "PeriodicStarMetric")

import warnings

import numpy as np
from scipy.optimize import curve_fit

from rubin_sim.maf.metrics.base_metric import BaseMetric
from rubin_sim.maf.utils import m52snr


class PeriodicStar:
    def __init__(self, filternames):
        self.filternames = filternames

    def __call__(self, t, x0, x1, x2, x3, x4, x5, x6, x7, x8):
        """Approximate a periodic star as a simple sin wave.
        t: array with "time" in days, and "filter" dtype names.
        x0: Period (days)
        x1: Phase (days)
        x2: Amplitude (mag)
        x3: mean u mag
        x4: mean g mag
        x5: mean r mag
        x6: mean i mag
        x7: mean z mag
        x8: mean y mag
        """
        filter2index = {"u": 3, "g": 4, "r": 5, "i": 6, "z": 7, "y": 8}
        filter_names = np.unique(self.filternames)
        mags = x2 * np.sin((t + x1) / x0 * 2.0 * np.pi)
        x = [x0, x1, x2, x3, x4, x5, x6, x7, x8]
        for f in filter_names:
            good = np.where(self.filternames == f)
            mags[good] += x[filter2index[f]]
        return mags


class PeriodicStarMetric(BaseMetric):
    """At each slice_point, run a Monte Carlo simulation to see how
    well a periodic source can be fit. Assumes a simple sin-wave light-curve,
    and generates Gaussain noise based in the 5-sigma limiting depth
    of each observation.

    Parameters
    ----------
    period : `float`
        The period to check, in days.
    amplitude : `float`
        The amplitude of the sinusoidal light curve, in mags.
    phase : `float`
        The phase of the lightcurve at the time of the first observation.
    n_monte : `int`
        The number of noise realizations to make in the Monte Carlo.
    period_tol : `float`
        The fractional tolerance on the period to require in order for a star
        to be considered well-fit
    amp_tol : `float`
        The fractional tolerance on the amplitude.
    means : `list` [`float`]
        The mean magnitudes in ugrizy of the star.
    mag_tol : `float`
        The mean magnitude tolerance, in magnitudes, for the star to be
        considered well-fit.
    n_bands : `int`
        Number of bands that must be within mag_tol.
    seed : `int`
        Random number seed for the noise realizations.
    """

    def __init__(
        self,
        metric_name="PeriodicStarMetric",
        mjd_col="observationStartMJD",
        m5_col="fiveSigmaDepth",
        filter_col="filter",
        period=10.0,
        amplitude=0.5,
        phase=2.0,
        n_monte=1000,
        period_tol=0.05,
        amp_tol=0.10,
        means=[20.0, 20.0, 20.0, 20.0, 20.0, 20.0],
        mag_tol=0.10,
        n_bands=3,
        seed=42,
        **kwargs,
    ):
        """
        period: days (default 10)
        amplitude: mags (default 1)
        n_monte: number of noise realizations to make in the Monte Carlo
        period_tol: fractional tolerance on the period to demand for a star
        to be considered well-fit
        amp_tol: fractional tolerance on the amplitude to demand
        means: mean magnitudes for ugrizy
        mag_tol: Mean magnitude tolerance (mags)
        n_bands: Number of bands that must be within mag_tol
        seed: random number seed
        """
        self.mjd_col = mjd_col
        self.m5_col = m5_col
        self.filter_col = filter_col
        super(PeriodicStarMetric, self).__init__(
            col=[self.mjd_col, self.m5_col, self.filter_col],
            units="Fraction Detected",
            metric_name=metric_name,
            **kwargs,
        )
        self.period = period
        self.amplitude = amplitude
        self.phase = phase
        self.n_monte = n_monte
        self.period_tol = period_tol
        self.amp_tol = amp_tol
        self.means = np.array(means)
        self.mag_tol = mag_tol
        self.n_bands = n_bands
        np.random.seed(seed)
        self.filter2index = {"u": 3, "g": 4, "r": 5, "i": 6, "z": 7, "y": 8}

    def run(self, data_slice, slice_point=None):
        # Bail if we don't have enough points
        # (need to fit mean magnitudes in each of the available bands -
        # self.means and for a period, amplitude, and phase)
        if data_slice.size < self.means.size + 3:
            return self.badval

        # Generate input for true light curve
        t = np.empty(data_slice.size, dtype=list(zip(["time", "filter"], [float, "|U1"])))
        t["time"] = data_slice[self.mjd_col] - data_slice[self.mjd_col].min()
        t["filter"] = data_slice[self.filter_col]

        # If we are adding a distance modulus to the magnitudes
        if "distMod" in list(slice_point.keys()):
            mags = self.means + slice_point["distMod"]
        else:
            mags = self.means
        true_params = np.append(np.array([self.period, self.phase, self.amplitude]), mags)
        true_obj = PeriodicStar(t["filter"])
        true_lc = true_obj(t["time"], *true_params)

        # Array to hold the fit results
        fits = np.zeros((self.n_monte, true_params.size), dtype=float)
        for i in np.arange(self.n_monte):
            snr = m52snr(true_lc, data_slice[self.m5_col])
            dmag = 2.5 * np.log10(1.0 + 1.0 / snr)
            noise = np.random.randn(true_lc.size) * dmag
            # Suppress warnings about failing on covariance
            fit_obj = PeriodicStar(t["filter"])
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                # If it fails to converge, save values that should fail later
                try:
                    parm_vals, pcov = curve_fit(
                        fit_obj, t["time"], true_lc + noise, p0=true_params, sigma=dmag
                    )
                except RuntimeError:
                    parm_vals = true_params * 0 + np.inf
            fits[i, :] = parm_vals

        # Throw out any magnitude fits if there are no observations
        # in that filter
        ufilters = np.unique(data_slice[self.filter_col])
        if ufilters.size < 9:
            for key in list(self.filter2index.keys()):
                if key not in ufilters:
                    fits[:, self.filter2index[key]] = -np.inf

        # Find the fraction of fits that meet the "well-fit" criteria
        period_frac_err = np.abs((fits[:, 0] - true_params[0]) / true_params[0])
        amp_frac_err = np.abs((fits[:, 2] - true_params[2]) / true_params[2])
        mag_err = np.abs(fits[:, 3:] - true_params[3:])
        n_bands = np.zeros(mag_err.shape, dtype=int)
        n_bands[np.where(mag_err <= self.mag_tol)] = 1
        n_bands = np.sum(n_bands, axis=1)
        n_recovered = np.size(
            np.where(
                (period_frac_err <= self.period_tol)
                & (amp_frac_err <= self.amp_tol)
                & (n_bands >= self.n_bands)
            )[0]
        )
        frac_recovered = float(n_recovered) / self.n_monte

        return frac_recovered



__all__ = [
    "predicted_zeropoint",
    "predicted_zeropoint_hardware",
    "predicted_zeropoint_itl",
    "predicted_zeropoint_hardware_itl",
    "predicted_zeropoint_e2v",
    "predicted_zeropoint_hardware_e2v",
]

import numpy as np


def predicted_zeropoint(band: str, airmass: float, exptime: float = 1) -> float:
    """General zeropoint values derived from v1.9 throughputs.

    Extinction coefficients and zeropoint intercepts calculated in
    https://github.com/lsst-pst/syseng_throughputs/blob/main/notebooks/InterpolateZeropoint.ipynb

    Parameters
    ----------
    band : `str`
        The bandpass name.
    airmass : `float`
        The airmass at which to evaluate the zeropoint.
    exptime : `float`, optional
        The exposure time to calculate zeropoint.

    Returns
    -------
    zeropoint : `float`
        The zeropoint for the `band` at airmass `airmass` for an exposure
        of time `exptime`.

    Notes
    -----
    Useful for comparing to DM pipeline zeropoints or for calculating counts
    in an image (`np.power(10, (zp - mag)/2.5)) = counts`).
    """
    extinction_coeff = {
        "u": -0.45815823969467745,
        "g": -0.20789273881603035,
        "r": -0.12233514157672552,
        "i": -0.07387773563152214,
        "z": -0.0573260392897174,
        "y": -0.09549137502152871,
    }
    # Interestingly, because these come from a fit over X, these
    # values are not identical to the zp values found for one second,
    # without considering the fit. They are within 0.005 magnitudes though.
    zeropoint_X1 = {
        "u": 26.52229760811932,
        "g": 28.50754554409417,
        "r": 28.360365503331952,
        "i": 28.170373076693625,
        "z": 27.781368851776005,
        "y": 26.813708013594344,
    }

    return zeropoint_X1[band] + extinction_coeff[band] * (airmass - 1) + 2.5 * np.log10(exptime)


def predicted_zeropoint_hardware(band: str, exptime: float = 1) -> float:
    """Zeropoint values for the hardware throughput curves only,
    without atmospheric contributions.

    Parameters
    ----------
    band : `str`
        The bandpass name.
    exptime : `float`, optional
        The exposure time to calculate zeropoint.

    Returns
    -------
    zeropoint : `float`
        The zeropoint for the hardware only component of `band` for an
        exposure of `exptime`.

    Notes
    -----
    This hardware-only zeropoint is primarily useful for converting sky
    background magnitudes into counts.
    """
    zeropoint = {
        "u": 26.99435242519598,
        "g": 28.72132437054738,
        "r": 28.487206668180864,
        "i": 28.267160381353793,
        "z": 27.850681356053688,
        "y": 26.988827459758397,
    }
    return zeropoint[band] + 2.5 * np.log10(exptime)


def predicted_zeropoint_itl(band: str, airmass: float, exptime: float = 1) -> float:
    """Average ITL zeropoint values derived from v1.9 throughputs.

    Parameters
    ----------
    band : `str`
        The bandpass name.
    airmass : `float`
        The airmass at which to evaluate the zeropoint.
    exptime : `float`, optional
        The exposure time to calculate zeropoint.

    Returns
    -------
    zeropoint : `float`
        The zeropoint for the `band` at airmass `airmass` for an exposure
        of time `exptime`.

    Notes
    -----
    Useful for comparing to DM pipeline zeropoints or for calculating counts
    in an image (`np.power(10, (zp - mag)/2.5)) = counts`).
    """
    extinction_coeff = {
        "u": -0.45815223255080606,
        "g": -0.20789245761381037,
        "r": -0.12233512060667238,
        "i": -0.07387767800064417,
        "z": -0.05739100372986528,
        "y": -0.09474605376660676,
    }
    zeropoint_X1 = {
        "u": 26.52231025834397,
        "g": 28.507547620761827,
        "r": 28.360365812523426,
        "i": 28.170380673108845,
        "z": 27.796728189989665,
        "y": 26.870922441512732,
    }

    return zeropoint_X1[band] + extinction_coeff[band] * (airmass - 1) + 2.5 * np.log10(exptime)


def predicted_zeropoint_hardware_itl(band: str, exptime: float = 1) -> float:
    """Zeropoint values for the ITL hardware throughput curves only,
    without atmospheric contributions.

    Parameters
    ----------
    band : `str`
        The bandpass name.
    exptime : `float`, optional
        The exposure time to calculate zeropoint.

    Returns
    -------
    zeropoint : `float`
        The zeropoint for the hardware only component of `band` for an
        exposure of `exptime`.

    Notes
    -----
    This hardware-only zeropoint is primarily useful for converting sky
    background magnitudes into counts.
    """
    zeropoint = {
        "u": 26.994361876151476,
        "g": 28.721326283280213,
        "r": 28.487206961551088,
        "i": 28.26716792192087,
        "z": 27.86618994188104,
        "y": 27.04446387553851,
    }
    return zeropoint[band] + 2.5 * np.log10(exptime)


def predicted_zeropoint_e2v(band: str, airmass: float, exptime: float = 1) -> float:
    """Average E2V zeropoint values derived from v1.9 throughputs.

    Parameters
    ----------
    band : `str`
        The bandpass name.
    airmass : `float`
        The airmass at which to evaluate the zeropoint.
    exptime : `float`, optional
        The exposure time to calculate zeropoint.

    Returns
    -------
    zeropoint : `float`
        The zeropoint for the `band` at airmass `airmass` for an exposure
        of time `exptime`.

    Notes
    -----
    Useful for comparing to DM pipeline zeropoints or for calculating counts
    in an image (`np.power(10, (zp - mag)/2.5)) = counts`).
    """
    extinction_coeff = {
        "u": -0.4600735940453953,
        "g": -0.20651340321330425,
        "r": -0.12276192263131014,
        "i": -0.07398443681400534,
        "z": -0.057334002964289726,
        "y": -0.095496483868828,
    }
    zeropoint_X1 = {
        "u": 26.58989678516041,
        "g": 28.567959743207357,
        "r": 28.44712188941494,
        "i": 28.19470048013101,
        "z": 27.7817595301949,
        "y": 26.813791858927964,
    }

    return zeropoint_X1[band] + extinction_coeff[band] * (airmass - 1) + 2.5 * np.log10(exptime)


def predicted_zeropoint_hardware_e2v(band: str, exptime: float = 1) -> float:
    """Zeropoint values for the E2V hardware throughput curves only,
    without atmospheric contributions.

    Parameters
    ----------
    band : `str`
        The bandpass name.
    exptime : `float`, optional
        The exposure time to calculate zeropoint.

    Returns
    -------
    zeropoint : `float`
        The zeropoint for the hardware only component of `band` for an
        exposure of `exptime`.

    Notes
    -----
    This hardware-only zeropoint is primarily useful for converting sky
    background magnitudes into counts.
    """
    zeropoint = {
        "u": 27.063967445283826,
        "g": 28.78030646345493,
        "r": 28.574328242939043,
        "i": 28.291563456601306,
        "z": 27.85108207854988,
        "y": 26.988912028019346,
    }

    return zeropoint[band] + 2.5 * np.log10(exptime)



# An extension to the GalaxyCountsMetric from Lynne Jones:
# maf_contrib/mafContrib/lssMetrics.py
#
# Purpose: Estimate the number of galaxies expected at a particular
# coadded depth, accounting for
# dust extinction, magnitude cuts, as well as redshift-bin-specific
# powerlaws (based on mock catalogs
# from Nelson D. Padilla et al.).
#
# Includes functionality to calculate the galaxy counts from CFHTLS power
# law from LSST Science Book
# as well as to normalize the galaxy counts from mock catalogs to match
# those with CFHTLS power law at i<25.5.
#
# Need constantsForPipeline.py to import the power law constants and
# the normalization factor.
#
# Humna Awan: humna.awan@rutgers.edu

__all__ = ("GalaxyCountsMetricExtended",)

import warnings

import numpy as np
import scipy

from rubin_sim.maf.maf_contrib.lss_obs_strategy.constants_for_pipeline import (
    normalization_constant,
    power_law_const_a,
    power_law_const_b,
)
from rubin_sim.maf.metrics import BaseMetric, Coaddm5Metric, ExgalM5


class GalaxyCountsMetricExtended(BaseMetric):
    """Estimate galaxy counts per HEALpix pixel.

    Accommodates for dust extinction, magnitude cuts,
    and specification of the galaxy LF to specific redshift bin to consider.
    Dependency (aside from MAF): constantsForPipeline.py

    Parameters
    ------------
    m5_col : `str`
        name of column for depth in the data. Default: 'fiveSigmaDepth'
    nside : `int`, opt
        HEALpix resolution parameter. Default: 128
    upper_mag_limit : `float`
        upper limit on magnitude when calculating the galaxy counts.
        Default: 32.0
    include_dust_extinction : `bool`
        set to False if do not want to include dust extinction.
        Default: True
    filter_band : `str`, opt
        any one of 'u', 'g', 'r', 'i', 'z', 'y'. Default: 'i'
    redshift_bin : `str`, opt
        options include '0.<z<0.15', '0.15<z<0.37', '0.37<z<0.66,
        '0.66<z<1.0',
        '1.0<z<1.5', '1.5<z<2.0', '2.0<z<2.5', '2.5<z<3.0','3.0<z<3.5',
        '3.5<z<4.0',
        'all' for no redshift restriction (i.e. 0.<z<4.0)
        Default: 'all'
    cfht_ls_counts : `bool`, opt
        set to True if want to calculate the total galaxy counts from CFHTLS
        powerlaw from LSST Science Book. Must be run with redshift_bin= 'all'
        Default: False
    normalized_mock_catalog_counts : `bool`, opt
     set to False if  want the raw/un-normalized galaxy counts from
     mock catalogs.
     Default: True
    """

    def __init__(
        self,
        m5_col="fiveSigmaDepth",
        filter_col="filter",
        nside=128,
        metric_name="GalaxyCountsMetricExtended",
        units="Galaxy Counts",
        upper_mag_limit=32.0,
        include_dust_extinction=True,
        filter_band="i",
        redshift_bin="all",
        cfht_ls_counts=False,
        normalized_mock_catalog_counts=True,
        **kwargs,
    ):
        self.m5_col = m5_col
        self.filter_col = filter_col
        self.upper_mag_limit = upper_mag_limit
        self.include_dust_extinction = include_dust_extinction
        self.redshift_bin = redshift_bin
        self.filter_band = filter_band
        self.cfhtls_counts = cfht_ls_counts
        self.normalized_mock_catalog_counts = normalized_mock_catalog_counts
        # Use the coadded depth metric to calculate the coadded depth
        # at each point.
        # Specific band (e.g. r-band) will be provided by the sql constraint.
        if self.include_dust_extinction:
            # include dust extinction when calculating the co-added depth
            self.coaddmetric = ExgalM5(m5_col=self.m5_col)
        else:
            self.coaddmetric = Coaddm5Metric(m5_col=self.m5_col)

        # Need to scale down to indivdual HEALpix pixels.
        # Galaxy count from the coadded depth is per 1 square degree.
        # Number of galaxies ~= 41253 sq. degrees in the full sky divided
        # by number of HEALpix pixels.
        self.scale = 41253.0 / (int(12) * nside**2)
        # Consider power laws from various redshift bins: importing
        # the constant
        # General power law form: 10**(a*m+b).
        self.power_law_const_a = power_law_const_a
        self.power_law_const_b = power_law_const_b

        super().__init__(
            col=[self.m5_col, self.filter_col],
            metric_name=metric_name,
            maps=self.coaddmetric.maps,
            units=units,
            **kwargs,
        )

    # ------------------------------------------------------------------------
    # set up the integrand to calculate galaxy counts
    def _gal_count(self, apparent_mag, coaddm5):
        # calculate the change in the power law constant based on the band
        # colors assumed here: (u-g)=(g-r)=(r-i)=(i-z)= (z-y)=0.4
        factor = 0.4
        band_correction_dict = {
            "u": -3.0 * factor,
            "g": -2.0 * factor,
            "r": -1.0 * factor,
            "i": 0.0,
            "z": factor,
            "y": 2.0 * factor,
        }
        if self.filter_band not in band_correction_dict:
            warnings.warn("Invalid band in GalaxyCountsMetricExtended. " "Assuming i-band instead.")
        band_correction = band_correction_dict.get(self.filter_band, 0.0)

        # check to make sure that the z-bin assigned is valid.
        if (self.redshift_bin != "all") and (self.redshift_bin not in list(self.power_law_const_a.keys())):
            warnings.warn(
                "Invalid redshift bin in GalaxyCountsMetricExtended. " "Defaulting to all redshifts."
            )
            self.redshift_bin = "all"

        # consider the power laws
        if self.redshift_bin == "all":
            if self.cfhtls_counts:
                # LSST power law: eq. 3.7 from LSST Science Book
                # converted to per sq degree:
                # (46*3600)*10^(0.31(i-25))
                dn_gal = 46.0 * 3600.0 * np.power(10.0, 0.31 * (apparent_mag + band_correction - 25.0))
            else:
                # full z-range considered here: 0.<z<4.0
                # sum the galaxy counts from each individual z-bin
                dn_gal = 0.0
                for key in list(self.power_law_const_a.keys()):
                    dn_gal += np.power(
                        10.0,
                        self.power_law_const_a[key] * (apparent_mag + band_correction)
                        + self.power_law_const_b[key],
                    )
        else:
            dn_gal = np.power(
                10.0,
                self.power_law_const_a[self.redshift_bin] * (apparent_mag + band_correction)
                + self.power_law_const_b[self.redshift_bin],
            )

        completeness = 0.5 * scipy.special.erfc(apparent_mag - coaddm5)
        return dn_gal * completeness

    # ------------------------------------------------------------------------
    def run(self, data_slice, slice_point=None):
        # Calculate the coadded depth.
        infilt = np.where(data_slice[self.filter_col] == self.filter_band)[0]
        # If there are no visits in this filter,
        # return immediately with a flagged value
        if len(infilt) == 0:
            return self.badval

        coaddm5 = self.coaddmetric.run(data_slice[infilt], slice_point)

        # some coaddm5 values are really small (i.e. min=10**-314).
        # Zero them out.
        if coaddm5 < 1:
            num_gal = 0

        else:
            num_gal, int_err = scipy.integrate.quad(
                self._gal_count, -np.inf, self.upper_mag_limit, args=coaddm5
            )
            # Normalize the galaxy counts (per sq deg)
            if self.normalized_mock_catalog_counts and not self.cfhtls_counts:
                num_gal = normalization_constant * num_gal
            if num_gal < 1.0:
                num_gal = 0.0
            # scale down to individual HEALpix pixel instead of per sq deg
            num_gal *= self.scale
        return num_gal



__all__ = ("PairMetric",)

import numpy as np

from .base_metric import BaseMetric


class PairMetric(BaseMetric):
    """Count the number of pairs of visits that could be used for
    Solar System object detection.

    Parameters
    ----------
    mjd_col : `str`, opt
        Name of the MJD column in the observations.
    metric_name : `str`, opt
        Name for the resulting metric. If None, one is constructed from
        the class name.
    match_min : `float`, opt
        Minutes after first observation to count something as a match.
    match_max : `float`, opt
        Minutes after first observation to count something as a match.
    bin_size : `float`, opt
        bin_size to use (minutes).
        Note that bin_size should be considerably smaller than the difference
        between match_min and match_max.

    Result
    ------
    num_pairs : `float`
        The number of pairs of visits within the min and max time range.
    """

    def __init__(
        self,
        mjd_col="observationStartMJD",
        metric_name="Pairs",
        match_min=20.0,
        match_max=40.0,
        bin_size=5.0,
        **kwargs,
    ):
        self.mjd_col = mjd_col
        self.bin_size = bin_size / 60.0 / 24.0
        self.match_min = match_min / 60.0 / 24.0
        self.match_max = match_max / 60.0 / 24.0
        super(PairMetric, self).__init__(col=mjd_col, metric_name=metric_name, units="N Pairs", **kwargs)

    def run(self, data_slice, slice_point=None):
        bins = np.arange(
            data_slice[self.mjd_col].min(),
            data_slice[self.mjd_col].max() + self.bin_size,
            self.bin_size,
        )

        hist, bin_edges = np.histogram(data_slice[self.mjd_col], bins=bins)
        nbin_min = np.round(self.match_min / self.bin_size)
        nbin_max = np.round(self.match_max / self.bin_size)
        bins_to_check = np.arange(nbin_min, nbin_max + 1, 1)
        bins_w_obs = np.where(hist > 0)[0]
        # now, for each bin with an observation,
        # need to check if there is a bin
        # far enough ahead that is also populated.
        result = 0
        for binadd in bins_to_check:
            result += np.size(np.intersect1d(bins_w_obs, bins_w_obs + binadd))
        if result == 0:
            result = self.badval
        return result



__all__ = ("stars_project", "assign_patches")

import numpy as np
from rubin_scheduler.utils import gnomonic_project_toxy


def stars_project(stars, visit):
    """
    Project the stars to x,y plane for a given visit.
    """
    xtemp, ytemp = gnomonic_project_toxy(
        np.radians(stars["ra"]),
        np.radians(stars["decl"]),
        np.radians(visit["ra"]),
        np.radians(visit["dec"]),
    )
    # Rotate the field using the visit rotSkyPos.
    # Hope I got that sign right...
    sin_rot = np.sin(np.radians(visit["rotSkyPos"]))
    cos_rot = np.cos(np.radians(visit["rotSkyPos"]))
    stars["x"] = cos_rot * xtemp + sin_rot * ytemp
    stars["y"] = -1.0 * sin_rot * xtemp + cos_rot * ytemp

    stars["radius"] = (stars["x"] ** 2 + stars["y"] ** 2) ** 0.5
    return stars


def assign_patches(stars, visit, n_patches=16, radius_fov=1.8):
    """
    Assign PatchIDs to everything.
    Assume that stars have already been projected to x,y
    """
    maxx, maxy = gnomonic_project_toxy(0.0, np.radians(radius_fov), 0.0, 0.0)
    nsides = n_patches**0.5

    # This should move all coords to  0 < x < nsides-1
    px = np.floor((stars["x"] + maxy) / (2.0 * maxy) * nsides)
    py = np.floor((stars["y"] + maxy) / (2.0 * maxy) * nsides)

    stars["sub_patch"] = px + py * nsides
    stars["patch_id"] = stars["sub_patch"] + visit["observationId"] * n_patches
    return stars



__all__ = ("CumulativeMetric",)

import numpy as np

from .base_metric import BaseMetric


class CumulativeMetric(BaseMetric):
    """For plotting up the cumulative number of observations.
    Expected to be used with a UniSlicer or UserPointSlicer with one point.

    Parameters
    ----------
    interp_points : `np.array`, (N,)  or None
        The points to interpolate the cumulative number of observations to.
        If None, then the range of the data is used with a stepsize of 1.
    """

    def __init__(
        self,
        metric_name="Cumulative",
        time_col="observationStartMJD",
        night_col="night",
        interp_points=None,
        **kwargs,
    ):
        super().__init__(col=[time_col, night_col], metric_name=metric_name, metric_dtype="object", **kwargs)
        self.time_col = time_col
        self.night_col = night_col
        self.interp_points = interp_points
        self.plot_dict = {"xlabel": "MJD (days)", "ylabel": "N obs"}

    def run(self, data_slice, slice_point=None):
        data_slice.sort(order=self.time_col)
        if self.interp_points is None:
            interp_points = np.arange(
                data_slice[self.night_col].min(),
                data_slice[self.night_col].max() + 1,
                1,
            )
        else:
            interp_points = self.interp_points
        cumulative_number = np.arange(data_slice.size) + 1
        yresult = np.interp(interp_points, data_slice[self.night_col], cumulative_number)
        xresult = interp_points
        return {"x": xresult, "y": yresult, "plot_dict": self.plot_dict}



"""Young Stellar Objects metric.
Converted from notebook 211116_yso_3D_90b.ipynb.
Formatted with black."""

__all__ = ("NYoungStarsMetric",)

import healpy as hp
import numpy as np
import scipy.integrate as integrate

from rubin_sim.maf.maps import DustMap, DustMap3D, StellarDensityMap
from rubin_sim.maf.metrics import BaseMetric, CrowdingM5Metric
from rubin_sim.phot_utils import DustValues


class StarDensity:
    """integrate from zero to some max distance, then multiply by angular area

    Parameters
    ----------
    l : float
        Galactic longitude, radians
    b : float
        Galactic latitude, radians
    """

    def __init__(self, gall, galb):
        """Calculate the expected number of stars along a line of site"""
        self.r_thin = 2.6  # scale length of the thin disk, kpc
        self.d_gc = 8.178  # Distance to the galactic center, kpc
        self.h_thin = 0.300  # scale height of the thin disk, kpc

        self.gall = gall
        self.galb = galb

        self.A = 0.8e8 / (4.0 * np.pi * self.h_thin * self.r_thin**2)

    def __call__(self, r):
        """
        Parameters
        ----------
        r : float
            Distance in kpc
        """
        r_galac = ((self.d_gc - r * np.cos(self.gall)) ** 2 + (r * np.sin(self.gall)) ** 2) ** 0.5

        exponent = -1.0 * r * np.abs(np.sin(self.galb)) / self.h_thin - r_galac / self.r_thin

        result = self.A * r**2 * np.exp(exponent)
        return result


class NYoungStarsMetric(BaseMetric):
    """Calculate the distance or number of stars with
    color uncertainty defined by mags/snrs.

    Parameters
    ----------
    metric_name : `str`, opt
        Default 'young_stars'.
    m5_col : `str`, opt
        The default column name for m5 information in the input data.
        Default fiveSigmaDepth.
    filter_col : `str`, opt
        The column name for the filter information. Default filter.
    mags : `dict`, opt
        The absolute magnitude of the object in question.
        Keys of filter name, values in mags.
        Default is for a 0.3 solar mass star at age = 100 Myr.
    snrs : `dict`, opt
        The SNR to demand for each filter.
    galb_limit : `float`, opt
        The galactic latitude above which to return zero (degrees).
        Default 90.
    badval : `float`, opt
        The value to return when the metric value cannot be calculated.
        Default 0.
    return_distance : `bool`, opt
        Whether the metric will return the maximum distance that
        can be reached for each slice_point,
        or the total number of stars down to mags/snrs.
    crowding_error : `float`, opt
        Crowding error that gets passed to CrowdingM5Metric. Default 0.25.
    use_2D_extinction : `bool`, opt
        Uses the 2D extinction map instead of the 3D one. Default False.
    """

    def __init__(
        self,
        metric_name="young_stars",
        m5_col="fiveSigmaDepth",
        filter_col="filter",
        seeing_col="seeingFwhmGeom",
        mags={"g": 10.32, "r": 9.28, "i": 7.97},
        snrs={"g": 5.0, "r": 5.0, "i": 5.0},
        galb_limit=90.0,
        badval=0,
        return_distance=False,
        crowding_error=0.25,
        use_2D_extinction=False,
        **kwargs,
    ):
        cols = [m5_col, filter_col, seeing_col]
        maps = [
            DustMap3D(),
            StellarDensityMap(filtername="g"),
            StellarDensityMap(filtername="r"),
            StellarDensityMap(filtername="i"),
        ]
        self.use_2D_extinction = use_2D_extinction
        if self.use_2D_extinction:
            maps[0] = DustMap()
        # This will give us access to the dust map get_distance_at_dmag routine
        # but does not require loading another copy of the map
        self.ebvmap = DustMap3D()
        self.return_distance = return_distance
        units = "kpc" if self.return_distance else "N stars"
        super().__init__(cols, metric_name=metric_name, maps=maps, units=units, badval=badval, **kwargs)
        # Save R_x values for on-the-fly calculation of dust extinction
        self.r_x = DustValues().r_x.copy()
        # set return type
        self.m5_col = m5_col
        self.filter_col = filter_col
        self.galb_limit = np.radians(galb_limit)
        self.mags = mags
        self.filters = list(self.mags.keys())
        self.snrs = snrs
        self.crowding_error = crowding_error
        self.m5crowding = {
            f: CrowdingM5Metric(
                crowding_error=crowding_error,
                filtername=f,
                seeing_col=seeing_col,
                maps=maps,
            )
            for f in self.filters
        }
        dust_properties = DustValues()
        self.ax1 = dust_properties.ax1

    def run(self, data_slice, slice_point=None):
        # Evaluate area on sky for this slice_point, in radians
        if "nside" in slice_point:
            # Best area /pixel calculation, appropriate for healpix slicers
            sky_area = hp.nside2pixarea(slice_point["nside"], degrees=False)
        else:
            # Assume single, approximate, circular FOV
            sky_area = np.pi * (np.radians(1.75)) ** 2

        # if we are outside the galb_limit, return nothing
        # Note we could make this a more complicated function that
        # returns an expected density of star forming regions
        if np.abs(slice_point["galb"]) > self.galb_limit:
            return self.badval

        # Compute depth for each filter
        depths = {}
        # ignore the divide by zero warnings
        with np.errstate(divide="ignore"):
            for filtername in self.filters:
                in_filt = np.where(data_slice[self.filter_col] == filtername)[0]
                # Calculate coadded depth per filter
                depth_m5 = 1.25 * np.log10(np.sum(10.0 ** (0.8 * data_slice[self.m5_col][in_filt])))
                if self.crowding_error > 0:
                    depth_crowding = self.m5crowding[filtername].run(data_slice, slice_point)
                    depths[filtername] = min(depth_m5, depth_crowding)
                else:
                    depths[filtername] = depth_m5

        # solve for the distances in each filter where we hit the required SNR
        distances = []
        for filtername in self.filters:
            # Apparent magnitude at the SNR requirement
            m_app = -2.5 * np.log10(self.snrs[filtername] / 5.0)
            m_app += depths[filtername]
            if self.use_2D_extinction:
                A_x = self.ax1[filtername] * slice_point["ebv"]
                # Assuming all the dust along the line of sight matters.
                m_app = m_app - A_x
                dist = 10.0 * (100 ** ((m_app - self.mags[filtername]) / 5.0)) ** 0.5
            else:
                dist = self.ebvmap.distance_at_dmag(
                    dmag=m_app - self.mags[filtername],
                    dists=slice_point["ebv3d_dists"],
                    ebvs=slice_point["ebv3d_ebvs"],
                    filtername=filtername,
                )
            distances.append(dist)
        # compute the final distance, limited by whichever filter is
        # most shallow
        final_distance = np.min(distances, axis=-1) / 1e3  # to kpc
        if self.return_distance:
            return final_distance
        else:
            # Resorting to numerical integration of ugly function
            sd = StarDensity(slice_point["gall"], slice_point["galb"])
            stars_per_sterr, _err = integrate.quad(sd, 0, final_distance)
            stars_tot = stars_per_sterr * sky_area
            return stars_tot



import unittest

import numpy as np

import rubin_sim.maf.metrics as metrics


class TestHourglassmetric(unittest.TestCase):
    def test_hourglass_metric(self):
        """Test the hourglass metric"""
        names = ["observationStartMJD", "night", "filter"]
        types = [float, float, str]
        npts = 50
        data = np.zeros(npts, dtype=list(zip(names, types)))
        day0 = 59000
        data["observationStartMJD"] = np.arange(0, 10, 0.2)[:npts] + day0
        data["night"] = np.floor(data["observationStartMJD"] - day0)
        data["filter"] = "r"
        data["filter"][-1] = "g"
        slice_point = [0]
        metric = metrics.HourglassMetric()
        result = metric.run(data, slice_point)
        pernight = result["pernight"]
        perfilter = result["perfilter"]

        # All the gaps are larger than 2 min.
        assert np.size(perfilter) == 2 * data.size
        # Check that the format is right at least
        assert len(pernight.dtype.names) == 9


if __name__ == "__main__":
    unittest.main()



__all__ = ("WFDlabelStacker",)

import healpy as hp
import numpy as np

from .base_stacker import BaseStacker


class WFDlabelStacker(BaseStacker):
    """Add an 'areaId' column to flag whether a visit was
    inside the 'footprint'.

    Parameters
    ----------
    footprint : `np.NDarray`, optional
        The healpix map indicating the desired footprint region.
        If this is not defined (default None),
        then the entire sky is used as the footprint.
    fp_threshold : `float`, optional
        The threshold for the fraction of the visit area which
        falls within the footprint in order
        to be counted as 'in' the footprint. Default 0.4.
    areaIdName : `str`, optional
        Value to place in the areaId column for visits which match this area.
    ra_col : `str`, optional
        The name of the RA column. Default fieldRA.
    dec_col : `str`, optional
        The name of the Dec column. Default fieldDec.
    note_col : `str`, optional
        The name of the 'note' (scheduler_note) column in the database.
        This is used to identify visits
        which were part of a DD sequence.
    exclude_dd : `bool`, optional
        Exclude (True) or include (False) visits which are part of a
        DD sequence within this 'area'.

    Notes
    -----
    This stacker adds an areaId column in the opsim database,
    to be labelled with 'areaIdName' if the visit falls within the
    healpix footprint map and (optionally) is not tagged as a DD visit.
    If it falls outside the footprint, the visit is tagged as "NULL".
    If it was part of a DD sequence, the visit is tagged with an
    ID which is unique to that DD field, if 'exclude_dd' is True.
    Generally this would be likely to be used to tag visits as
    belonging to WFD - but not necessarily!
    Any healpix footprint is valid.
    """

    cols_added = ["area_id"]

    def __init__(
        self,
        footprint=None,
        fp_threshold=0.4,
        area_id_name="WFD",
        ra_col="fieldRA",
        dec_col="fieldDec",
        note_col="scheduler_note",
        exclude_dd=True,
    ):
        self.ra_col = ra_col
        self.dec_col = dec_col
        self.note_col = note_col
        self.cols_req = [self.ra_col, self.dec_col, self.note_col]
        self.cols_added_dtypes = [(str, 15)]
        self.units = [""]
        self.fp_threshold = fp_threshold
        self.area_id_name = area_id_name
        self.exclude_dd = exclude_dd
        if footprint is None:
            # If footprint was not defined,
            # just set it to cover the entire sky, at nside=64
            footprint = np.ones(hp.nside2npix(64))
        self.footprint = footprint
        self.nside = hp.npix2nside(len(self.footprint))

    def define_ddname(self, note):
        field = note.replace("u,", "")
        field = field.split(",")[0].replace(",", "")
        return field

    def _run(self, sim_data, cols_present=False):
        # Even if cols_present is true, recalculate.
        # Set up DD names.
        d = set()
        for p in np.unique(sim_data[self.note_col]):
            if p.startswith("DD"):
                d.add(self.define_ddname(p))
        # Identify Healpixels associated with each visit.
        vec = hp.dir2vec(sim_data[self.ra_col], sim_data[self.dec_col], lonlat=True)
        vec = vec.swapaxes(0, 1)
        radius = np.radians(1.75)  # fov radius
        area_id = np.zeros(len(sim_data), self.cols_added_dtypes[0])
        for i, (v, note) in enumerate(zip(vec, sim_data[self.note_col])):
            # Identify the healpixels which would be inside this pointing
            pointing_healpix = hp.query_disc(self.nside, v, radius, inclusive=False)
            # The wfd_footprint consists of values of 0/1
            # if out/in WFD footprint
            hp_in_fp = self.footprint[pointing_healpix].sum()
            # So in_fp= the number of healpixels which were in
            # the specified footprint
            # .. in the # in / total # > limit (0.4)
            # then "yes" it's in the footprint
            in_fp = hp_in_fp / len(pointing_healpix)
            if note.startswith("DD") and self.exclude_dd:
                area_id[i] = self.define_ddname(note)
            else:
                if in_fp >= self.fp_threshold:
                    area_id[i] = self.area_id_name
                else:
                    area_id[i] = "NULL"
        sim_data["area_id"] = area_id
        return sim_data



__all__ = ("MafRunResults",)

import os
import re
from collections import OrderedDict

import numpy as np

import rubin_sim.maf.db as db
import rubin_sim.maf.metric_bundles as metricBundles


class MafRunResults:
    """Read and serve the MAF resultsDb_sqlite.db database for the
    show_maf jinja2 templates.

    Deals with a single MAF run (one output directory, one results_db) only.

    Parameters
    ----------
    out_dir : `str`
        The location of the results database for this run.
    run_name : `str`, optional
        The name of the opsim run.
        If None, simply stays blank on show_maf display pages.
    results_db : `str`, optional
        The path to the sqlite database in `out_dir`.
        If None, uses the default of `resultsDb_sqlite.db`.
    """

    def __init__(self, out_dir, run_name=None, results_db=None):
        self.out_dir = os.path.relpath(out_dir, ".")
        self.run_name = run_name

        # Read in the results database.
        if results_db is None:
            results_db = os.path.join(self.out_dir, "resultsDb_sqlite.db")
        database = db.ResultsDb(database=results_db)

        # Get the metric and display info (1-1 match)
        self.metrics = database.get_metric_display_info()
        self.metrics = self.sort_metrics(self.metrics)

        # Get the plot and stats info (many-1 metric match)
        skip_stats = ["Completeness@Time", "Completeness H", "FractionPop "]
        self.stats = database.get_summary_stats(summary_name_notlike=skip_stats)
        self.plots = database.get_plot_files()

        # Pull up the names of the groups and subgroups.
        groups = sorted(np.unique(self.metrics["display_group"]))
        self.groups = OrderedDict()
        for g in groups:
            group_metrics = self.metrics[np.where(self.metrics["display_group"] == g)]
            self.groups[g] = sorted(np.unique(group_metrics["display_subgroup"]))

        self.summary_stat_order = [
            "Id",
            "Identity",
            "Median",
            "Mean",
            "Rms",
            "RobustRms",
            "N(-3Sigma)",
            "N(+3Sigma)",
            "Count",
            "25th%ile",
            "75th%ile",
            "Min",
            "Max",
        ]
        # Add in the table fraction sorting to summary stat ordering.
        table_fractions = [
            x for x in list(np.unique(self.stats["summary_metric"])) if x.startswith("TableFraction")
        ]
        if len(table_fractions) > 0:
            for x in (
                "TableFraction 0 == P",
                "TableFraction 1 == P",
                "TableFraction 1 < P",
            ):
                if x in table_fractions:
                    table_fractions.remove(x)
            table_fractions = sorted(table_fractions)
            self.summary_stat_order.append("TableFraction 0 == P")
            for table_frac in table_fractions:
                self.summary_stat_order.append(table_frac)
            self.summary_stat_order.append("TableFraction 1 == P")
            self.summary_stat_order.append("TableFraction 1 < P")

        self.plot_order = ["SkyMap", "Histogram", "PowerSpectrum", "Combo"]

    # Methods to deal with metricIds

    def convert_select_to_metrics(self, group_list, metric_id_list):
        """
        Convert the lists of values returned by 'select metrics' template page
        into an appropriate dataframe of metrics (in sorted order).

        Parameters
        ----------
        group_list : `list` [`str`]
            The groups of metrics to show on the show_maf pages.
        metric_id_list : `list` [`int`]
            The integer ids of the metrics in the sqlite results database.

        Returns
        -------
        metrics : `np.ndarray`, (N,)
            An array of the metric information for the metrics .
        """
        metric_ids = set()
        for group_subgroup in group_list:
            group = group_subgroup.split("_")[0]
            subgroup = group_subgroup.split("_")[-1].replace("+", " ")
            m_ids = self.metric_ids_in_subgroup(group, subgroup)
            for m_id in m_ids:
                metric_ids.add(m_id)
        for m_id in metric_id_list:
            m_id = int(m_id)
            metric_ids.add(m_id)
        metric_ids = list(metric_ids)
        metrics = self.metric_ids_to_metrics(metric_ids)
        metrics = self.sort_metrics(metrics)
        return metrics

    def get_json(self, metric):
        """
        Return the JSON string containing the data for a particular metric.
        """
        if len(metric) > 1:
            return None
        metric = metric[0]
        filename = metric["metric_datafile"]
        if filename.upper() == "NULL":
            return None
        datafile = os.path.join(self.out_dir, filename)
        # Read data back into a  bundle.
        m_b = metricBundles.create_empty_metric_bundle()
        m_b.read(datafile)
        io = m_b.output_json()
        if io is None:
            return None
        return io.get_value()

    def get_npz(self, metric):
        """
        Return the npz data.
        """
        if len(metric) > 1:
            return None
        metric = metric[0]
        filename = metric["metric_datafile"]
        if filename.upper() == "NULL":
            return None
        else:
            datafile = os.path.join(self.out_dir, filename)
            return datafile

    def get_results_db(self):
        """
        Return the summary results sqlite filename, as long as the
        results data is named `resultsDb_sqlite.db`.
        """
        return os.path.join(self.out_dir, "resultsDb_sqlite.db")

    def metric_ids_in_subgroup(self, group, subgroup):
        """
        Return the metric_ids within a given group/subgroup.
        """
        metrics = self.metrics_in_subgroup(group, subgroup)
        metric_ids = list(metrics["metric_id"])
        return metric_ids

    def metric_ids_to_metrics(self, metric_ids, metrics=None):
        """
        Return an ordered numpy array of metrics matching metric_ids.
        """
        if metrics is None:
            metrics = self.metrics
        metrics = metrics[np.isin(metrics["metric_id"], metric_ids)]
        return metrics

    def metrics_to_metric_ids(self, metrics):
        """
        Return a list of the metric Ids corresponding to a subset of metrics.
        """
        return list(metrics["metric_id"])

    # Methods to deal with metrics in numpy recarray.

    def sort_metrics(
        self,
        metrics,
        order=(
            "display_group",
            "display_subgroup",
            "base_metric_name",
            "slicer_name",
            "display_order",
            "metric_info_label",
        ),
    ):
        """
        Sort the metrics by order specified by 'order'.

        Default is to sort by group, subgroup, metric name, slicer,
        display order, then info_label.
        Returns sorted numpy array.
        """
        if len(metrics) > 0:
            metrics = np.sort(metrics, order=order)
        return metrics

    def metrics_in_group(self, group, metrics=None, sort=True):
        """
        Given a group, return the metrics belonging to this group,
        in display order.
        """
        if metrics is None:
            metrics = self.metrics
        metrics = metrics[np.where(metrics["display_group"] == group)]
        if sort:
            metrics = self.sort_metrics(metrics)
        return metrics

    def metrics_in_subgroup(self, group, subgroup, metrics=None):
        """
        Given a group and subgroup, return a dataframe of the metrics
        belonging to these group/subgroups, in display order.

        If 'metrics' is provided, then only consider this subset of metrics.
        """
        metrics = self.metrics_in_group(group, metrics, sort=False)
        if len(metrics) > 0:
            metrics = metrics[np.where(metrics["display_subgroup"] == subgroup)]
            metrics = self.sort_metrics(metrics)
        return metrics

    def metrics_to_subgroups(self, metrics):
        """
        Given an array of metrics, return an ordered dict of their
        group/subgroups.
        """
        group_list = sorted(np.unique(metrics["display_group"]))
        groups = OrderedDict()
        for group in group_list:
            groupmetrics = self.metrics_in_group(group, metrics, sort=False)
            groups[group] = sorted(np.unique(groupmetrics["display_subgroup"]))
        return groups

    def metrics_with_plot_type(self, plot_type="SkyMap", metrics=None):
        """
        Return an array of metrics with plot=plot_type
        (optionally also within a metric subset).
        """
        # Allow some variation in plot_type names for backward compatibility,
        #  even if plot_type is  a list.
        if not isinstance(plot_type, list):
            plot_type = [plot_type]
        plot_types = []
        for p_t in plot_type:
            plot_types.append(p_t)
            if p_t.endswith("lot"):
                plot_types.append(p_t[:-4])
            else:
                plot_types.append(p_t.lower() + "Plot")
        if metrics is None:
            metrics = self.metrics
        # Identify the plots with the right plot_type, get their IDs.
        plot_match = self.plots[np.isin(self.plots["plot_type"], plot_types)]
        # Convert those potentially matching metricIds to metrics,
        # using the subset info.
        metrics = self.metric_ids_to_metrics(plot_match["metric_id"], metrics)
        return metrics

    def unique_metric_names(self, metrics=None, baseonly=True):
        """
        Return a list of the unique metric names,
        preserving the order of 'metrics'.
        """
        if metrics is None:
            metrics = self.metrics
        if baseonly:
            sort_name = "base_metric_name"
        else:
            sort_name = "metric_name"
        metric_names = list(np.unique(metrics[sort_name]))
        return metric_names

    def metrics_with_summary_stat(self, summary_stat_name="Identity", metrics=None):
        """
        Return metrics with summary stat matching 'summary_stat_name'
        (optionally, within a metric subset).
        """
        if metrics is None:
            metrics = self.metrics
        # Identify the potentially matching stats.
        stats = self.stats[np.isin(self.stats["summary_metric"], summary_stat_name)]
        # Identify the subset of relevant metrics.
        metrics = self.metric_ids_to_metrics(stats["metric_id"], metrics)
        # Re-sort metrics because at this point, probably want displayOrder
        # + info_label before metric name.
        metrics = self.sort_metrics(
            metrics,
            order=[
                "display_group",
                "display_subgroup",
                "slicer_name",
                "display_order",
                "metric_info_label",
                "base_metric_name",
            ],
        )
        return metrics

    def metrics_with_stats(self, metrics=None):
        """
        Return metrics that have any summary stat.
        """
        if metrics is None:
            metrics = self.metrics
        # Identify metricIds which are also in stats.
        metrics = metrics[np.isin(metrics["metric_id"], self.stats["metric_id"])]
        metrics = self.sort_metrics(
            metrics,
            order=[
                "display_group",
                "display_subgroup",
                "slicer_name",
                "display_order",
                "metric_info_label",
                "base_metric_name",
            ],
        )
        return metrics

    def unique_slicer_names(self, metrics=None):
        """
        For an array of metrics, return the unique slicer names.
        """
        if metrics is None:
            metrics = self.metrics
        return list(np.unique(metrics["slicer_name"]))

    def metrics_with_slicer(self, slicer, metrics=None):
        """
        For an array of metrics, return the subset which match a
        particular 'slicername' value.
        """
        if metrics is None:
            metrics = self.metrics
        metrics = metrics[np.where(metrics["slicer_name"] == slicer)]
        return metrics

    def unique_metric_name_and_info_label(self, metrics=None):
        """
        For an array of metrics, return the unique metric names
        + info_label combo in same order.
        """
        if metrics is None:
            metrics = self.metrics
        metric_info_label = []
        for metric_name, info_label in zip(metrics["metric_name"], metrics["metric_info_label"]):
            metricinfo = " ".join([metric_name, info_label])
            if metricinfo not in metric_info_label:
                metric_info_label.append(metricinfo)
        return metric_info_label

    def unique_metric_info_label(self, metrics=None):
        """
        For an array of metrics, return a list of the unique info_label.
        """
        if metrics is None:
            metrics = self.metrics
        return list(np.unique(metrics["metric_info_label"]))

    def metrics_with_info_label(self, info_label, metrics=None):
        """
        For an array of metrics, return the subset which match a
        particular 'info_label' value.
        """
        if metrics is None:
            metrics = self.metrics
        metrics = metrics[np.where(metrics["metric_info_label"] == info_label)]
        return metrics

    def metrics_with_metric_name(self, metric_name, metrics=None, baseonly=True):
        """
        Return all metrics which match metric_name
        (default, only the 'base' metric name).
        """
        if metrics is None:
            metrics = self.metrics
        if baseonly:
            metrics = metrics[np.where(metrics["base_metric_name"] == metric_name)]
        else:
            metrics = metrics[np.where(metrics["metric_name"] == metric_name)]
        return metrics

    def metric_info(self, metric=None, with_data_link=False, with_slicer_name=True):
        """
        Return a dict with the metric info we want to show on the webpages.

        Currently : MetricName / Slicer/ InfoLabel / datafile (for download)
        Used to build a lot of tables in showMaf.
        """
        metric_info = OrderedDict()
        if metric is None:
            metric_info["Metric Name"] = ""
            if with_slicer_name:
                metric_info["Slicer"] = ""
            metric_info["Info Label"] = ""
            if with_data_link:
                metric_info["Data"] = []
                metric_info["Data"].append([None, None])
            return metric_info
        # Otherwise, do this for real (not a blank).
        metric_info["Metric Name"] = metric["metric_name"]
        if with_slicer_name:
            metric_info["Slicer"] = metric["slicer_name"]
        metric_info["InfoL abel"] = metric["metric_info_label"]
        if with_data_link:
            metric_info["Data"] = []
            metric_info["Data"].append(metric["metric_datafile"])
            metric_info["Data"].append(os.path.join(self.out_dir, metric["metric_datafile"]))
        return metric_info

    def caption_for_metric(self, metric):
        """
        Return the caption for a given metric.
        """
        caption = metric["display_caption"]
        if caption == "NULL":
            return ""
        else:
            return caption

    # Methods for plots.

    def plots_for_metric(self, metric):
        """
        Return a numpy array of the plots which match a given metric.
        """
        return self.plots[np.where(self.plots["metric_id"] == metric["metric_id"])]

    def plot_dict(self, plots=None):
        """
        Given an array of plots (for a single metric usually).
        Returns an ordered dict with 'plot_type' for interfacing with
        jinja2 templates.
        plot_dict ==
        {'SkyMap': {'plot_file': [], 'thumb_file', []}, 'Histogram': {}..}

        If no plot of a particular type, the plot_file and thumb_file
        are empty lists.
        Calling with plots=None returns a blank plot_dict.
        """
        plot_dict = OrderedDict()
        # Go through plots in 'plotOrder'.
        if plots is None:
            for p in self.plot_order:
                plot_dict[p] = {}
                plot_dict[p]["plot_file"] = ""
                plot_dict[p]["thumb_file"] = ""
        else:
            plot_types = list(np.unique(plots["plot_type"]))
            for p in self.plot_order:
                if p in plot_types:
                    plot_dict[p] = {}
                    plotmatch = plots[np.where(plots["plot_type"] == p)]
                    plot_dict[p]["plot_file"] = []
                    plot_dict[p]["thumb_file"] = []
                    for pm in plotmatch:
                        plot_dict[p]["plot_file"].append(self.get_plot_file(pm))
                        plot_dict[p]["thumb_file"].append(self.get_thumb_file(pm))
                    plot_types.remove(p)
            # Round up remaining plots.
            for p in plot_types:
                plot_dict[p] = {}
                plotmatch = plots[np.where(plots["plot_type"] == p)]
                plot_dict[p]["plot_file"] = []
                plot_dict[p]["thumb_file"] = []
                for pm in plotmatch:
                    plot_dict[p]["plot_file"].append(self.get_plot_file(pm))
                    plot_dict[p]["thumb_file"].append(self.get_thumb_file(pm))
        return plot_dict

    def get_thumb_file(self, plot):
        """
        Return the thumbnail file name for a given plot.
        """
        thumb_file = os.path.join(self.out_dir, plot["thumb_file"])
        return thumb_file

    def get_plot_file(self, plot):
        """
        Return the filename for a given plot.
        """
        plot_file = os.path.join(self.out_dir, plot["plot_file"])
        return plot_file

    def order_plots(self, sky_plots):
        """
        sky_plots = numpy array of skymap plots.

        Returns an ordered list of plotDicts.

        The goal is to lay out the skymaps in a 3x2 grid on the MultiColor
        page, in ugrizy order.
        If a plot for a filter is missing, add a gap. (i.e. if there is no
        u band plot, keep a blank spot).
        If there are other plots, with multiple filters or no filter
        info, they are added to the end.
        If sky_plots includes multiple plots in the same filter,
        just goes back to displayOrder.
        """
        ordered_sky_plots = []
        if len(sky_plots) == 0:
            return ordered_sky_plots

        order_list = ["u", "g", "r", "i", "z", "y"]
        blank_plot_dict = self.plot_dict(None)

        # Look for filter names in the plot filenames.
        too_many_plots = False
        for f in order_list:
            pattern = "_" + f + "_"
            matches = np.array([bool(re.search(pattern, x)) for x in sky_plots["plot_file"]])
            match_sky_plot = sky_plots[matches]
            if len(match_sky_plot) == 1:
                ordered_sky_plots.append(self.plot_dict(match_sky_plot))
            elif len(match_sky_plot) == 0:
                ordered_sky_plots.append(blank_plot_dict)
            else:
                # If we found more than one plot in the same filter,
                # we just go back to displayOrder.
                too_many_plots = True
                break

        if too_many_plots is False:
            # Add on any additional non-filter plots (e.g. joint completeness)
            #  that do NOT match original _*_ pattern.
            pattern = "_[ugrizy]_"
            nonmatches = np.array([bool(re.search(pattern, x)) for x in sky_plots["plot_file"]])
            nonmatch_sky_plots = sky_plots[nonmatches == False]
            if len(nonmatch_sky_plots) > 0:
                for sky_plot in nonmatch_sky_plots:
                    ordered_sky_plots.append(self.plot_dict(np.array([sky_plot])))

        elif too_many_plots:
            metrics = self.metrics[np.isin(self.metrics["metric_id"], sky_plots["metric_id"])]
            metrics = self.sort_metrics(metrics, order=["display_order"])
            ordered_sky_plots = []
            for m in metrics:
                sky_plot = sky_plots[np.where(sky_plots["metric_id"] == m["metric_id"])]
                ordered_sky_plots.append(self.plot_dict(sky_plot))

        # Pad out to make sure there are rows of 3
        while len(ordered_sky_plots) % 3 != 0:
            ordered_sky_plots.append(blank_plot_dict)

        return ordered_sky_plots

    def get_sky_maps(self, metrics=None, plot_type="SkyMap"):
        """
        Return a numpy array of the plots with plot_type=plot_type,
        optionally for subset of metrics.
        """
        if metrics is None:
            metrics = self.metrics
        # Match the plots to the metrics required.
        plot_metric_match = self.plots[np.isin(self.plots["metric_id"], metrics["metric_id"])]
        # Match the plot type (which could be a list)
        plot_match = plot_metric_match[np.isin(plot_metric_match["plot_type"], plot_type)]
        return plot_match

    # Set of methods to deal with summary stats.

    def stats_for_metric(self, metric, stat_name=None):
        """
        Return a numpy array of summary statistics which match a
        given metric(s).

        Optionally specify a particular stat_name that you want to match.
        """
        stats = self.stats[np.where(self.stats["metric_id"] == metric["metric_id"])]
        if stat_name is not None:
            stats = stats[np.where(stats["summary_metric"] == stat_name)]
        return stats

    def stat_dict(self, stats):
        """
        Returns an ordered dictionary with statName:statValue
        for an array of stats.

        Note that if you pass 'stats' from multiple metrics with the same
        summary names, they will be overwritten in the resulting dictionary!
        So just use stats from one metric, with unique summary_metric names.
        """
        # Result = dict with key
        # == summary stat name, value = summary stat value.
        sdict = OrderedDict()
        statnames = self.order_stat_names(stats)
        for n in statnames:
            match = stats[np.where(stats["summary_metric"] == n)]
            # We're only going to look at the first value;
            # and this should be a float.
            sdict[n] = match["summary_value"][0]
        return sdict

    def order_stat_names(self, stats):
        """
        Given an array of stats, return a list containing all the unique
        'summary_metric' names in a default ordering
        (identity-count-mean-median-rms..).
        """
        names = list(np.unique(stats["summary_metric"]))
        # Add some default sorting:
        namelist = []
        for nord in self.summary_stat_order:
            if nord in names:
                namelist.append(nord)
                names.remove(nord)
        for remaining in names:
            namelist.append(remaining)
        return namelist

    def all_stat_names(self, metrics):
        """
        Given an array of metrics, return a list containing all the
        unique 'summary_metric' names in a default ordering.
        """
        names = np.unique(
            self.stats["summary_metric"][np.isin(self.stats["metric_id"], metrics["metric_id"])]
        )
        names = list(names)
        # Add some default sorting.
        namelist = []
        for nord in self.summary_stat_order:
            if nord in names:
                namelist.append(nord)
                names.remove(nord)
        for remaining in names:
            namelist.append(remaining)
        return namelist



__all__ = ("ChebyValues",)

import os

import numpy as np
import pandas as pd

from .chebyshev_utils import chebeval


class ChebyValues:
    """Calculates positions, velocities, deltas, vmags and elongations,
    given a series of coefficients generated by ChebyFits.
    """

    def __init__(self):
        self.coeffs = {}
        self.coeff_keys = [
            "obj_id",
            "t_start",
            "t_end",
            "ra",
            "dec",
            "geo_dist",
            "vmag",
            "elongation",
        ]
        self.ephemeris_keys = [
            "ra",
            "dradt",
            "dec",
            "ddecdt",
            "geo_dist",
            "vmag",
            "elongation",
        ]

    def set_coefficients(self, cheby_fits):
        """Set coefficients using a ChebyFits object.
        (which contains a dictionary of obj_id, t_start, t_end, ra,
        dec, delta, vmag, and elongation lists).

        Parameters
        ----------
        cheby_fits : `rubin_sim.movingObjects.chebyFits`
            ChebyFits object, with attribute 'coeffs' -
            a dictionary of lists of coefficients.
        """
        self.coeffs = cheby_fits.coeffs
        # Convert list of coefficients into numpy arrays.
        for k in self.coeffs:
            self.coeffs[k] = np.array(self.coeffs[k])
        # Check that expected values were received.
        missing_keys = set(self.coeff_keys) - set(self.coeffs)
        if len(missing_keys) > 0:
            raise ValueError("Expected to find key(s) %s in coefficients." % " ".join(list[missing_keys]))
        self.coeffs["meanRA"] = self.coeffs["ra"].swapaxes(0, 1)[0]
        self.coeffs["meanDec"] = self.coeffs["dec"].swapaxes(0, 1)[0]

    def read_coefficients(self, cheby_fits_file):
        """Read coefficients from output file written by ChebyFits.

        Parameters
        ----------
        cheby_fits_file : `str`
            The filename of the coefficients file.
        """
        if not os.path.isfile(cheby_fits_file):
            raise IOError("Could not find cheby_fits_file at %s" % (cheby_fits_file))
        # Read the coefficients file.
        coeffs = pd.read_table(cheby_fits_file, sep=r"\s+")
        # The header line provides information on the number of
        # coefficients for each parameter.
        datacols = coeffs.columns.values
        cols = {}
        coeff_cols = ["ra", "dec", "geo_dist", "vmag", "elongation"]
        for k in coeff_cols:
            cols[k] = [x for x in datacols if x.startswith(k)]
        # Translate dataframe to dictionary of numpy arrays
        # while consolidating RA/Dec/Delta/Vmag/Elongation coeffs.
        self.coeffs["obj_id"] = coeffs.obj_id.values
        self.coeffs["t_start"] = coeffs.t_start.values
        self.coeffs["t_end"] = coeffs.t_end.values
        for k in coeff_cols:
            self.coeffs[k] = np.empty([len(cols[k]), len(coeffs)], float)
            for i in range(len(cols[k])):
                self.coeffs[k][i] = coeffs["%s_%d" % (k, i)].values
        # Add the mean RA and Dec columns
        # (before swapping the coefficients axes).
        self.coeffs["meanRA"] = self.coeffs["ra"][0]
        self.coeffs["meanDec"] = self.coeffs["dec"][0]
        # Swap the coefficient axes so that they are [segment, coeff].
        for k in coeff_cols:
            self.coeffs[k] = self.coeffs[k].swapaxes(0, 1)

    def _eval_segment(self, segment_idx, times, subset_segments=None, mask=True):
        """Evaluate the ra/dec/delta/vmag/elongation values for a
        given segment at a series of times.

        Parameters
        ----------
        segment_idx : `int`
            The index in (each of) self.coeffs for the segment.
            e.g. the first segment, for each object.
        times : `np.ndarray`
            The times at which to evaluate the segment.
        subset_segments : `np.ndarray`, optional
            Optionally specify a subset of the total segment indexes.
            This lets you pick out particular obj_ids.
        mask : `bool`, optional
            If True, returns NaNs for values outside the range of times
            in the segment.
            If False, extrapolates segment for times outside the
            segment time range.

        Returns
        -------
        ephemeris : `dict`
           Dictionary of RA, Dec, delta, vmag, and elongation values for
           the segment indicated, at the time indicated.
        """
        if subset_segments is None:
            subset_segments = np.ones(len(self.coeffs["obj_id"]), dtype=bool)
        t_start = self.coeffs["t_start"][subset_segments][segment_idx]
        t_end = self.coeffs["t_end"][subset_segments][segment_idx]
        t_scaled = times - t_start
        t_interval = np.array([t_start, t_end]) - t_start
        # Evaluate RA/Dec/Delta/Vmag/elongation.
        ephemeris = {}
        ephemeris["ra"], ephemeris["dradt"] = chebeval(
            t_scaled,
            self.coeffs["ra"][subset_segments][segment_idx],
            interval=t_interval,
            do_velocity=True,
            mask=mask,
        )
        ephemeris["dec"], ephemeris["ddecdt"] = chebeval(
            t_scaled,
            self.coeffs["dec"][subset_segments][segment_idx],
            interval=t_interval,
            do_velocity=True,
            mask=mask,
        )
        ephemeris["dradt"] = ephemeris["dradt"] * np.cos(np.radians(ephemeris["dec"]))
        for k in ("geo_dist", "vmag", "elongation"):
            ephemeris[k], _ = chebeval(
                t_scaled,
                self.coeffs[k][subset_segments][segment_idx],
                interval=t_interval,
                do_velocity=False,
                mask=mask,
            )
        return ephemeris

    def get_ephemerides(self, times, obj_ids=None, extrapolate=False):
        """Find the ephemeris information for 'obj_ids' at 'time'.

        Implicit in how this is currently written is that the segments
        are all expected to cover the same start/end time range across
        all objects.
        They do not have to have the same segment length for all objects.

        Parameters
        ----------
        times : `float` or `np.ndarray`
            The time to calculate ephemeris positions.
        obj_ids : `np.ndarray`, opt
            The object ids for which to generate ephemerides.
            If None, then just uses all objects.
        extrapolate : `bool`, opt
            If True, extrapolate beyond ends of segments if time
            outside of segment range.
            If False, return ValueError if time is beyond range of segments.

        Returns
        -------
        ephemerides : `np.ndarray`
            The ephemeris positions for all objects.
            Note that these may not be sorted in the same order as obj_ids.
        """
        if isinstance(times, float) or isinstance(times, int):
            times = np.array([times], float)
        ntimes = len(times)
        ephemerides = {}
        # Find subset of segments which match obj_id, if specified.
        if obj_ids is None:
            obj_match = np.ones(len(self.coeffs["obj_id"]), dtype=bool)
            ephemerides["obj_id"] = np.unique(self.coeffs["obj_id"])
        else:
            if isinstance(obj_ids, str) or isinstance(obj_ids, int):
                obj_ids = np.array([obj_ids])
            obj_match = np.isin(self.coeffs["obj_id"], obj_ids)
            ephemerides["obj_id"] = obj_ids
        # Now find ephemeris values.
        ephemerides["time"] = np.zeros((len(ephemerides["obj_id"]), ntimes), float) + times
        for k in self.ephemeris_keys:
            ephemerides[k] = np.zeros((len(ephemerides["obj_id"]), ntimes), float)
        for it, t in enumerate(times):
            # Find subset of segments which contain the appropriate time.
            # Look for simplest subset first.
            segments = np.where(
                (self.coeffs["t_start"][obj_match] <= t) & (self.coeffs["t_end"][obj_match] > t)
            )[0]
            if len(segments) == 0:
                seg_start = self.coeffs["t_start"][obj_match].min()
                seg_end = self.coeffs["t_end"][obj_match].max()
                if seg_start > t or seg_end < t:
                    if not extrapolate:
                        for k in self.ephemeris_keys:
                            ephemerides[k][:, it] = np.nan
                    else:
                        # Find the segments to use to extrapolate the times.
                        if seg_start > t:
                            segments = np.where(self.coeffs["t_start"][obj_match] == seg_start)[0]
                        if seg_end < t:
                            segments = np.where(self.coeffs["t_end"][obj_match] == seg_end)[0]
                elif seg_end == t:
                    # Not extrapolating, but outside the
                    # simple match case above.
                    segments = np.where(self.coeffs["t_end"][obj_match] == seg_end)[0]
            for i, segmentIdx in enumerate(segments):
                ephemeris = self._eval_segment(segmentIdx, t, obj_match, mask=False)
                for k in self.ephemeris_keys:
                    ephemerides[k][i][it] = ephemeris[k]
                ephemerides["obj_id"][i] = self.coeffs["obj_id"][obj_match][segmentIdx]
        if obj_ids is not None:
            if set(ephemerides["obj_id"]) != set(obj_ids):
                raise ValueError(
                    "Did not find expected match between obj_ids provided and ephemeride obj_ids."
                )
        return ephemerides



__all__ = ("combine_result_dbs", "gather_summaries")

import argparse
import glob
import os
import sqlite3

import numpy as np
import pandas as pd


def dirname_to_runname(inpath, replaces=["_glance", "_sci", "_meta", "_ss", "_ddf"]):
    """Given a directory path, construct a plausible runname"""
    result = os.path.basename(os.path.normpath(inpath))
    for rstring in replaces:
        result = result.replace(rstring, "")
    return result


def combine_result_dbs(run_dirs, dbfilename="resultsDb_sqlite.db"):
    """Helper function for gather_summaries

    Parameters
    ----------
    run_dirs : list of str
        A list of directories to search for MAF result databases.
    dbfilename : str (resultsDb_sqlite.db)
        The database filename to look for (default: resultsDb_sqlite.db)."""

    db_files = []
    run_names = []
    for dname in run_dirs:
        fname = os.path.join(dname, dbfilename)
        if os.path.isfile(fname):
            db_files.append(fname)
            run_names.append(dirname_to_runname(dname))

    # querry to grab all the summary stats
    sql_q = "SELECT summarystats.summary_value, "
    sql_q += "metrics.metric_name, metrics.metric_info_label, "
    sql_q += "metrics.slicer_name, summarystats.summary_name "
    sql_q += "FROM summarystats INNER JOIN metrics ON metrics.metric_id=summarystats.metric_id"

    rows = []
    for row_name, fname in zip(run_names, db_files):
        con = sqlite3.connect(fname)
        temp_df = pd.read_sql(sql_q, con)
        con.close()

        # Make column names
        col_names = []
        for summary_name, metric_name, metric_info_label, slicer_name in zip(
            temp_df["summary_name"].values.tolist(),
            temp_df["metric_name"].values.tolist(),
            temp_df["metric_info_label"].values.tolist(),
            temp_df["slicer_name"].values.tolist(),
        ):
            col_name = " ".join(
                [
                    summary_name.strip(),
                    metric_name.strip(),
                    metric_info_label.strip(),
                    slicer_name.strip(),
                ]
            )
            col_names.append(col_name.replace("  ", " "))

        # Make a DataFrame row
        row = pd.DataFrame(
            temp_df["summary_value"].values.reshape([1, temp_df["summary_value"].values.size]),
            columns=col_names,
            index=[row_name],
        )

        # Can have duplicate columns if MAF was run multiple times.
        # Remove duplicates:
        # https://stackoverflow.com/questions/14984119/
        # python-pandas-remove-duplicate-columns
        row = row.loc[:, ~row.columns.duplicated()].copy()

        rows.append(row)
    # Create final large DataFrame to hold everything
    all_cols = np.unique(np.concatenate([r.columns.values for r in rows]))
    u_names = np.unique(run_names)
    result_df = pd.DataFrame(
        np.zeros([u_names.size, all_cols.size]) + np.nan,
        columns=all_cols,
        index=u_names,
    )

    # Put each row into the final DataFrame
    for row_name, row in zip(run_names, rows):
        result_df.loc[row_name, row.columns] = np.ravel(row.values)

    return result_df


def gather_summaries():
    """Find resultsDbs in a series of directories and gather up their summary
    stats into a single CSV or hdf5 file. Outputs one row per unique run name.
    """

    parser = argparse.ArgumentParser(
        description="Find resultsDbs in a series of directories and "
        "gather up their summary stats into a single CSV or hdf5 file. "
        "Intended to run on a set of metrics run on multiple "
        "simulations, so that each results_db has similar summary"
        "statistics."
    )
    parser.add_argument(
        "--base_dir",
        type=str,
        default=".",
        help="Root directory from where to search for MAF (sub)directories.",
    )
    parser.add_argument(
        "--outfile",
        type=str,
        default="summary",
        help="Output file name. Default (summary)",
    )
    parser.add_argument(
        "--to_hdf",
        dest="to_hdf",
        action="store_true",
        help="Create a .hdf5 file, instead of the default csv file.",
    )
    parser.add_argument(
        "--dirs",
        type=str,
        default=None,
        help="comma seperated list of directories to use, default None",
    )

    args = parser.parse_args()
    if args.dirs is None:
        run_dirs = glob.glob(args.base_dir + "/*/")
    else:
        run_dirs = args.dirs.split(",")

    # Create output file name if needed
    if args.to_hdf:
        outfile = args.outfile + ".h5"
    else:
        outfile = args.outfile + ".csv"

    result_df = combine_result_dbs(run_dirs)

    # Save summary statistics
    if args.to_hdf:
        result_df.to_hdf(outfile, key="stats")
    else:
        # Create a CSV file
        result_df.to_csv(outfile)



__all__ = (
    "make_circle_subset_slicer",
    "make_wfd_subset_slicer",
    "HealpixSubsetSlicer",
)

from functools import wraps

import healpy as hp
import numpy as np
import rubin_scheduler.utils as simsUtils
from rubin_scheduler.scheduler.utils import get_current_footprint

from .healpix_slicer import HealpixSlicer


def make_circle_subset_slicer(ra_cen, dec_cen, radius=3.0, nside=512, use_cache=False):
    """Create a circular healpix subset slicer, centered on ra_cen/dec_cen.

    Parameters
    ----------
    ra_cen : `float`
        RA of the center of the slicer (degrees).
    dec_cen : `float`
        Dec of the center of the slicer (degrees).
    radius : `float`, optional
        Radius of the circular slicer shape (degrees).
    nside : `int`, optional
        Nside resolution of the healpix subset slicer.
    use_cache : `bool`, optional
        Set up the slicer to `use_cache` or not.

    Returns
    -------
    subsetslicer, plot_dict : `maf.HealpixSubsetSlicer`, `dict`
        A healpix subset slicer, defined according to a circle centered
        on `ra_cen`, `dec_cen`, with radius `radius`.
        Also returns a minimal plot dict, with the visufunc and rotation
        information to plot a small circular area with the HealpixSkyMap
        plotter.
    """
    ra, dec = simsUtils.hpid2_ra_dec(nside, np.arange(hp.nside2npix(nside)))
    dist = simsUtils.angular_separation(ra_cen, dec_cen, ra, dec)
    close = np.where(dist <= radius)[0]
    subsetslicer = HealpixSubsetSlicer(nside, close, use_cache=use_cache)
    plot_dict = {
        "visufunc": hp.gnomview,
        "rot": (ra_cen, dec_cen, 0),
        "xsize": 500,
    }
    return subsetslicer, plot_dict


def make_wfd_subset_slicer(nside=64, use_cache=True, wfd_labels=None):
    """Create a wfd-only healpix subset slicer.

    Parameters
    ----------
    nside : `int`, optional
        Nside resolution of the healpix subset slicer.
    use_cache : `bool`, optional
        Set up the slicer to `use_cache` or not.
    wfd_labels : `list` [`str`] or None
        List of the labels from the footprint to use for the "WFD"
        (or other) footprint. Default of None will use the current
        WFD level areas, including the WFD-level galactic plane area.

    Returns
    -------
    subsetslicer : `maf.HealpixSubsetSlicer`
        A healpix subset slicer, defined according to the most current
        version of the scheduler footprint using `get_current_footprint`.
    """
    if wfd_labels is None:
        wfd_labels = ["lowdust", "euclid_overlap", "virgo", "bulgy", "LMC_SMC"]
    footprints, labels = get_current_footprint(nside=nside)
    wfdpix = np.where(np.isin(labels, wfd_labels))[0]
    slicer = HealpixSubsetSlicer(nside=nside, hpid=wfdpix, use_cache=use_cache)
    return slicer


class HealpixSubsetSlicer(HealpixSlicer):
    """A spatial slicer that evaluates pointings on a subset of a healpix grid.

    The advantage of using this healpixSubsetSlicer (rather than just putting
    the RA/Dec values into the UserPointsSlicer, which is another valid
    approach) is that you preserve the full healpix array.
    This means you could do things like calculate the power spectrum
    and plot without remapping into healpixels first. The downside is that
    you must first (externally) define the healpixels that you
    wish to use.

    When plotting with RA/Dec, the default HealpixSkyMap can be used,
    corresponding to {'rot': (0, 0, 0), 'flip': 'astro'}.

    Parameters
    ----------
    nside : `int`
        The nside parameter of the healpix grid. Must be a power of 2.
    hpid : `np.ndarray`, (N,)
        The subset of healpix id's to use to calculate the metric.
        Because the hpid should be defined based on a particular nside,
        these first two arguments are not optional for this slicer.
    lon_col : `str`, optional
        Name of the longitude (RA equivalent) column to use
        from the input data. Default fieldRA
    lat_col : `str`, optional
        Name of the latitude (Dec equivalent) column to use from
        the input data. Default fieldDec
    lat_lon_deg : `bool`, optional
        Flag indicating whether the lat and lon values in the input data
        are in degrees (True) or radians (False).
        Default True.
    verbose : `bool`, optional
        Flag to indicate whether to write additional information
        to stdout during runtime. Default True.
    badval : `float`, optional
        Bad value flag, relevant for plotting.
        Default the np.nan value (in order to properly flag bad data points
        for plotting with the healpix plotting routines).
        In general, this should not be changed.
    use_cache : `bool`, optional
        Flag allowing the user to indicate whether to cache (and reuse)
        metric results calculated with the same set of simulated data
        pointings. This can be safely set to True for slicers not using
        maps and will result in increased speed.
        When calculating metric results using maps, the map data at each
        individual ra/dec point may influence the metric results and so
        use_cache should be set to False.
        Default True.
    leafsize : `int`, optional
        Leafsize value for kdtree. Default 100.
    radius : `float`, optional
        Radius for matching in the kdtree.
        Equivalent to the radius of the FOV. Degrees.
        Default 1.75.
    use_camera : `bool`, optional
        Flag to indicate whether to use the LSST camera footprint or not.
        Default False.
    camera_footprint_file : `str`, optional
        Name of the camera footprint map to use. Can be None, which will
        use the default.
    rot_sky_pos_col_name : `str`, optional
        Name of the rotSkyPos column in the input  data.
        Only used if use_camera is True.
        Describes the orientation of the camera orientation
        compared to the sky. Default rotSkyPos.
    """

    def __init__(
        self,
        nside,
        hpid,
        lon_col="fieldRA",
        lat_col="fieldDec",
        lat_lon_deg=True,
        verbose=True,
        badval=np.nan,
        use_cache=True,
        leafsize=100,
        radius=2.45,
        use_camera=True,
        camera_footprint_file=None,
        rot_sky_pos_col_name="rotSkyPos",
    ):
        """Instantiate and set up healpix slicer object."""
        super().__init__(
            nside=nside,
            verbose=verbose,
            lon_col=lon_col,
            lat_col=lat_col,
            badval=badval,
            radius=radius,
            leafsize=leafsize,
            use_cache=use_cache,
            use_camera=use_camera,
            rot_sky_pos_col_name=rot_sky_pos_col_name,
            camera_footprint_file=camera_footprint_file,
            lat_lon_deg=lat_lon_deg,
        )
        self.hpid = hpid
        # add hpid to the slicer_init values
        self.slicer_init["hpid"] = hpid
        self.len_hpid = len(self.hpid)
        # Set up a mask for the metric values to use
        self.mask = np.ones(hp.nside2npix(self.nside), bool)
        self.mask[self.hpid] = False

    def __eq__(self, other_slicer):
        """Evaluate if two slicers are equivalent."""
        # If the two slicers are both healpix slicers, check nsides value.
        result = False
        if isinstance(other_slicer, HealpixSubsetSlicer):
            if other_slicer.nside == self.nside:
                if np.array_equal(other_slicer.hpid, self.hpid):
                    if other_slicer.lon_col == self.lon_col and other_slicer.lat_col == self.lat_col:
                        if other_slicer.radius == self.radius:
                            if other_slicer.use_camera == self.use_camera:
                                if other_slicer.rotSkyPosColName == self.rotSkyPosColName:
                                    if np.all(other_slicer.shape == self.shape):
                                        result = True
        return result

    def __iter__(self):
        """Iterate over the slices."""
        self.hpid_counter = 0
        return self

    def __next__(self):
        """Returns results of self._slice_sim_data when iterating over slicer.

        Results of self._slice_sim_data should be dictionary of
        {'idxs': the data indexes relevant for this slice of the slicer,
        'slice_point': the metadata for the slice_point, which always
        includes 'sid' key for ID of slice_point.}
        """
        if self.hpid_counter >= self.len_hpid:
            raise StopIteration
        # Set up 'current'
        islice = self.hpid[self.hpid_counter]
        # Set up 'next'
        self.hpid_counter += 1
        # Return 'current'
        return self._slice_sim_data(islice)

    def setup_slicer(self, sim_data, maps=None):
        """Use sim_data[self.lon_col] and sim_data[self.lat_col]
        (in radians) to set up KDTree.

        Parameters
        -----------
        sim_data : `numpy.ndarray`, (N,)
            The simulated data, including the location of each pointing.
        maps : `list` of `rubin_sim.maf.maps` objects, optional
            List of maps (such as dust extinction) that will run to build
            up additional metadata at each slice_point.
            This additional metadata is available to metrics via the
            slice_point dictionary.
            Default None.
        """
        super().setup_slicer(sim_data=sim_data, maps=maps)

        @wraps(self._slice_sim_data)
        def _slice_sim_data(islice):
            """Return indexes for relevant opsim data at slice_point
            (slice_point=lon_col/lat_col value .. usually ra/dec)."""
            # Subclass this method, just to make sure we return
            # no data for points not in self.hpid
            slice_point = {"sid": islice, "nside": self.nside}
            if islice not in self.hpid:
                indices = []
            else:
                sx, sy, sz = simsUtils._xyz_from_ra_dec(
                    self.slice_points["ra"][islice], self.slice_points["dec"][islice]
                )
                # Query against tree.
                indices = self.opsimtree.query_ball_point((sx, sy, sz), self.rad)
                if (self.use_camera) & (len(indices) > 0):
                    # Find the indices *of those indices* which fall in
                    # the camera footprint
                    camera_idx = self.camera(
                        self.slice_points["ra"][islice],
                        self.slice_points["dec"][islice],
                        self.data_ra[indices],
                        self.data_dec[indices],
                        self.data_rot[indices],
                    )
                    indices = np.array(indices)[camera_idx]
                # Loop through all the slice_point keys.
                # If the first dimension of slice_point[key] has
                # the same shape as the slicer, assume it is information
                # per slice_point.
                # Otherwise, pass the whole slice_point[key] information.
                # Useful for stellar LF maps
                # where we want to pass only the relevant LF and the bins
                # that go with it.
                for key in self.slice_points:
                    if len(np.shape(self.slice_points[key])) == 0:
                        keyShape = 0
                    else:
                        keyShape = np.shape(self.slice_points[key])[0]
                    if keyShape == self.nslice:
                        slice_point[key] = self.slice_points[key][islice]
                    else:
                        slice_point[key] = self.slice_points[key]

            return {"idxs": indices, "slice_point": slice_point}

        setattr(self, "_slice_sim_data", _slice_sim_data)



__all__ = ("ebv_3d_hp", "get_x_at_nearest_y")

import os
import warnings

import healpy as hp
import numpy as np
from astropy.io import fits
from rubin_scheduler.data import get_data_dir

from rubin_sim.maf.utils import radec2pix


def ebv_3d_hp(
    nside,
    map_file=None,
    ra=None,
    dec=None,
    pixels=None,
    interp=False,
):
    """Reads and saves a 3d dust extinction file from disk, return extinction
    at  specified points (ra/dec/ or pixels).

    Parameters
    ----------
    nside : `int`
        Healpixel resolution (2^x).
    map_file : `str`, opt
        Path to dust map file.
    ra : `np.ndarray` or `float`, opt
        RA (can take numpy array).
        Default None sets up healpix array of nside. Radians.
    dec : `np.ndarray` or `float`, opt
        Dec (can take numpy array).
        Default None set up healpix array of nside. Radians.
    pixels : `np.ndarray`, opt
        Healpixel IDs, to sub-select particular healpix points.
        Default uses all points.
        Easiest way to access healpix values.
        Note that the pixels in the healpix array MUST come from a h
        ealpix grid with the same nside as the ebv_3d_hp map.
        Using different nsides can potentially fail silently.
    interp : `bool`, opt
        Should returned values be interpolated (True)
        or just nearest neighbor (False).
        Default False.
    """
    if (ra is None) & (dec is None) & (pixels is None):
        raise RuntimeError("Need to set ra,dec or pixels.")

    # Set up path to map data
    if map_file is None:
        if nside == 64:
            map_name = "merged_ebv3d_nside64_defaults.fits"
            map_file = os.path.join(get_data_dir(), "maps/DustMaps3D", map_name)
        elif nside == 128:
            map_name = "merged_ebv3d_nside128_defaults.fits"
            map_file = os.path.join(get_data_dir(), "maps/DustMaps3D", map_name)
        else:
            raise Exception(
                f"map_file not specified, and nside {nside} not one of 64 or 128. "
                "Please specify a known map_file, as a basis for interpolation."
            )
    else:
        # Check if user specified map name but not full map path
        if not os.path.isfile(map_file):
            test_path = os.path.join(get_data_dir(), "maps/DustMaps3D", map_file)
            if os.path.isfile(test_path):
                map_file = test_path
    # Keep track of what nside and what map_file we're using
    if not hasattr(ebv_3d_hp, "map_file"):
        ebv_3d_hp.mapFile = map_file
        ebv_3d_hp.nside = nside

    # Do we need to re-read from disk?
    if (not hasattr(ebv_3d_hp, "ebvs")) | (not hasattr(ebv_3d_hp, "dists")) | (map_file != ebv_3d_hp.mapFile):
        ebv_3d_hp.mapFile = map_file
        ebv_3d_hp.nside = nside
        # Read map from disk
        hdul = fits.open(map_file)
        # hpids = hdul[0].data
        dists = hdul[1].data
        ebvs = hdul[2].data
        hdr = hdul[0].header
        # Close map file
        hdul.close()

        # Check additional healpix information from the header
        map_nside = hdr["NSIDE"]
        nested = hdr["NESTED"]
        if nside != map_nside:
            warnings.warn(
                f"Map nside {map_nside} did not match expected nside {nside}; "
                f"Will use nside from map data."
            )
            if pixels is not None:
                # We're just going to raise an exception here
                # because this could mean bad things.
                raise ValueError(
                    f"Map nside {map_nside} did not match expected nside {nside}, "
                    f"and pixels provided; this can potentially indicate a serious "
                    f"error. Make nsides match or specify ra/dec instead of pixels."
                )
            nside = map_nside
        # Nested healpix data will not match the healpix arrays
        # for the slicers (at this time)
        if nested:
            warnings.warn("Map has nested (not ring order) data; will reorder.")
            for i in np.arange(0, len(dists[0])):
                dists[:, i] = hp.reorder(dists[:, i], "nest", "ring")
                ebvs[:, i] = hp.reorder(ebvs[:, i], "nest", "ring")
        ebv_3d_hp.dists = dists
        ebv_3d_hp.ebvs = ebvs
        print(f"Read map {map_file} from disk")

    if pixels is not None:
        # Assume if pixels is set, then interp is irrelevant
        dists = ebv_3d_hp.dists[pixels]
        ebvs = ebv_3d_hp.ebvs[pixels]
    # BUT, if we were provided ra/dec then have to see if we should interpolate
    else:
        if interp:  # find interpolated values at given ra/dec
            distlen = len(ebv_3d_hp.dists[0])
            dists = np.zeros([len(ra), distlen], float)
            ebvs = np.zeros([len(ra), distlen], float)
            for i in np.arange(0, distlen):
                dists[:, i] = hp.get_interp_val(ebv_3d_hp.dists[:, i], np.pi / 2.0 - dec, ra)
                ebvs[:, i] = hp.get_interp_val(ebv_3d_hp.ebvs[:, i], np.pi / 2.0 - dec, ra)
        else:  # just fine nearest neighbor pixel value and return those
            pixels = radec2pix(nside, ra, dec)
            dists = ebv_3d_hp.dists[pixels]
            ebvs = ebv_3d_hp.ebvs[pixels]

    return dists, ebvs


def get_x_at_nearest_y(x, y, x_goal):
    """Given a goal x value, find y values at the closest x value.

    This could be used to fetch the nearest ebv value at a given distance,
    or the nearest distance to a given m-Mo value, for example.

    Parameters
    ----------
    x : `np.array`
        Can be either a map with x at each point in the map (2d array) or
        the x at a single point of the map (1d array)
    y : `np.array`
        Can be either a map with y at each point in the map (2d array) or
        the y at a single point of the map (1d array) -
        but should match x dimensionality
    x_goal : `float'
        The goal x value to look for the nearest y value

    Returns
    -------
    x_closest, y_closest : `np.array`, `np.array`
        1-d array of x and y (single value or over map).
    """
    if x.ndim == 2:
        idx_min = np.argmin(np.abs(x - x_goal), axis=1)
        i_expand = np.expand_dims(idx_min, axis=-1)
        x_closest = np.take_along_axis(x, i_expand, axis=-1).squeeze()
        y_closest = np.take_along_axis(y, i_expand, axis=-1).squeeze()
    else:
        idx_min = np.argmin(np.abs(x - x_goal))
        x_closest = x[idx_min]
        y_closest = y[idx_min]
    return x_closest, y_closest



__all__ = ("scimaf_dir",)

import argparse
import glob
import os
import shutil
import sqlite3
import warnings

import matplotlib
import pandas as pd

matplotlib.use("Agg")

from . import batches as batches
from . import db as db
from . import metricBundles as mmB


def scimaf_dir():
    """Run the science batch on all .db files in a directory."""

    parser = argparse.ArgumentParser()
    parser.add_argument("--db", type=str, default=None)
    parser.add_argument(
        "--no_clobber",
        dest="no_clobber",
        action="store_false",
        help="Do not remove existing directory outputs",
    )
    parser.set_defaults(no_long_micro=False)
    parser.add_argument("--limited", dest="limited", action="store_true")
    parser.set_defaults(limited=False)
    args = parser.parse_args()

    if args.db is None:
        db_files = glob.glob("*.db")
        db_files = [filename for filename in db_files if "trackingDb" not in filename]
    else:
        db_files = [args.db]
    run_names = [os.path.basename(name).replace(".db", "") for name in db_files]

    for filename, name in zip(db_files, run_names):
        out_dir = name + "_sci"

        # Grab the starting date for the Presto KNe metric
        try:
            con = sqlite3.connect(filename)
            mjd0_df = pd.read_sql("select min(observationStartMJD) from observations;", con)
            con.close()
            mjd0 = mjd0_df.values.min()
        # If this fails for any reason (aka schema change)
        except:  # noqa E722
            warnings.warn("Could not find survey start date for Presto KNe, setting mjd0=None.")
            mjd0 = None
        # Clobber output directory if it exists
        if not args.no_clobber:
            if os.path.isdir(out_dir):
                shutil.rmtree(out_dir)
        results_db = db.ResultsDb(out_dir=out_dir)
        # Set up the metricBundles
        if args.limited:
            bdict = batches.radar_limited(
                runName=name,
                mjd0=mjd0,
            )
        else:
            bdict = batches.science_radar_batch(
                runName=name,
                mjd0=mjd0,
            )
        # Run them, including generating plots
        group = mmB.MetricBundleGroup(
            bdict, filename, out_dir=out_dir, results_db=results_db, save_early=False
        )
        group.run_all(clear_memory=True, plot_now=True)
        results_db.close()
        db.add_run_to_database(
            out_dir,
            "trackingDb_sqlite.db",
            run_group=None,
            run_name=name,
            run_comment=None,
            maf_comment="ScienceRadar",
            db_file=name + ".db",
        )



__all__ = ("BaseObs",)

import os
import warnings

import numpy as np
from rubin_scheduler.data import get_data_dir
from rubin_scheduler.utils import LsstCameraFootprint, angular_separation

from rubin_sim.phot_utils import Bandpass, Sed

from .ooephemerides import PyOrbEphemerides


class BaseObs:
    """
    Base class to generate observations of a set of moving objects.

    Parameters
    ----------
    footPrint: `str`, optional
        Specify the footprint for the FOV.
        Options include "camera", "circle", "rectangle".
        'Camera' means use the actual LSST camera footprint
        (following a rough cut with a circular FOV).
        Default is camera FOV.
    r_fov : `float`, optional
        If footprint is "circular", this is the radius of the fov (in degrees).
        Default 1.75 degrees (only used for circular fov).
    x_tol : `float`, optional
        If footprint is "rectangular", this is half of the width
        of the (on-sky) fov in the RA direction (in degrees).
        Default 5 degrees.
    y_tol : `float`, optional
        If footprint is "rectangular", this is half of the width of
        the fov in Declination (in degrees).
        Default is 3 degrees
    eph_mode: `str`, optional
        Mode for ephemeris generation - nbody or 2body. Default is nbody.
    eph_type: `str`, optional
        Type of ephemerides to generate - full or basic.
        Full includes all values calculated by openorb;
        Basic includes a more basic set.
        Default is Basic.
    eph_file: `str` or None, optional
        The name of the planetary ephemerides file to use
        for ephemeris generation.
        Default (None) will use the default for PyOrbEphemerides.
    obs_code: `str`, optional
        Observatory code for ephemeris generation.
        Default is "I11" - Cerro Pachon.
    obs_time_col: `str`, optional
        Name of the time column in the obsData.
        Default 'observationStartMJD'.
    obs_time_scale: `str`, optional
        Type of timescale for MJD (TAI or UTC currently).
        Default TAI.
    seeing_col: `str`, optional
        Name of the seeing column in the obsData.
        Default 'seeingFwhmGeom'.
        This should be the geometric/physical seeing
        as it is used for the trailing loss calculation.
    visit_exp_time_col: `str`, optional
        Name of the visit exposure time column in the obsData.
        Default 'visitExposureTime'.
    obs_ra: `str`, optional
        Name of the RA column in the obsData. Default 'fieldRA'.
    obs_dec: `str`, optional
        Name of the Dec column in the obsData. Default 'fieldDec'.
    obs_rot_sky_pos: `str`, optional
        Name of the Rotator column in the obsData. Default 'rotSkyPos'.
    obs_degrees: `bool`, optional
        Whether the observational data is in degrees or radians.
        Default True (degrees).
    outfile_name : `str`, optional
        The output file name.
        Default is 'lsst_obs.dat'.
    obs_info : `str`, optional
        A string that captures provenance information about the observations.
        For example: 'baseline_v2.0_10yrs, years 0-5'
        or 'baseline2018a minus NES'
        Default ''.
    """

    def __init__(
        self,
        footprint="camera",
        r_fov=1.75,
        x_tol=5,
        y_tol=3,
        eph_mode="nbody",
        eph_type="basic",
        obs_code="I11",
        eph_file=None,
        obs_time_col="observationStartMJD",
        obs_time_scale="TAI",
        seeing_col="seeingFwhmGeom",
        visit_exp_time_col="visitExposureTime",
        obs_ra="fieldRA",
        obs_dec="fieldDec",
        obs_rot_sky_pos="rotSkyPos",
        obs_degrees=True,
        outfile_name="lsst_obs.dat",
        obs_info="",
        camera_footprint_file=None,
    ):
        # Strings relating to the names of columns in the observation metadata.
        self.obs_code = obs_code
        self.obs_time_col = obs_time_col
        self.obs_time_scale = obs_time_scale
        self.seeing_col = seeing_col
        self.visit_exp_time_col = visit_exp_time_col
        self.obs_ra = obs_ra
        self.obs_dec = obs_dec
        self.obs_rot_sky_pos = obs_rot_sky_pos
        self.obs_degrees = obs_degrees
        # Save a space for the standard object colors.
        self.colors = {}
        self.outfile_name = outfile_name
        # Values for identifying observations.
        self.footprint = footprint.lower()
        if self.footprint == "camera":
            self._setup_camera(camera_footprint_file=camera_footprint_file)
        self.r_fov = r_fov
        self.x_tol = x_tol
        self.y_tol = y_tol
        # Values for ephemeris generation.
        if eph_mode.lower() not in ("2body", "nbody"):
            raise ValueError("Ephemeris generation must be 2body or nbody.")
        self.eph_mode = eph_mode
        self.eph_type = eph_type
        self.eph_file = eph_file

        self._setup_info(obs_info=obs_info)

    def _setup_info(self, obs_info=""):
        """Generate a dict to record relevant settings"""
        info = {
            "obs_info": obs_info,
            "footprint": self.footprint,
            "eph_mode": self.eph_mode,
            "eph_type": self.eph_type,
        }
        if self.footprint == "circle":
            info["r_fov"] = self.r_fov
        if self.footprint == "rectangle":
            info["xtol"] = self.x_tol
            info["ytol"] = self.y_tol

        # convert to numpy array in anticipation of saving
        names = list(info.keys())
        types = [np.array(info[key]).dtype for key in names]
        self.info = np.zeros(1, dtype=list(zip(names, types)))
        for key in names:
            self.info[key] = info[key]

    def _setup_camera(self, camera_footprint_file=None):
        self.camera = LsstCameraFootprint(units="degrees", footprint_file=camera_footprint_file)

    def setup_ephemerides(self):
        """Initialize the ephemeris generator.
        Save the setup PyOrbEphemeris class.

        This uses the default engine, pyoorb -
        however this could be overwritten to use another generator.
        """
        self.ephems = PyOrbEphemerides(ephfile=self.eph_file)

    def generate_ephemerides(self, sso, times, eph_mode=None, eph_type=None):
        """Generate ephemerides for 'sso' at times 'times'
        (assuming MJDs, with timescale self.obs_time_scale).

        The default engine here is pyoorb, however other ephemeris generation
        could be used with a matching API to PyOrbEphemerides.

        The initialized pyoorb class (PyOrbEphemerides) is saved,
        to skip setup on subsequent calls.

        Parameters
        ----------
        sso : `rubin_sim.movingObjects.Orbits`
            Typically this will be a single object.
        times: `np.ndarray`
            The times at which to generate ephemerides. MJD.
        eph_mode: `str` or None, optional
            Potentially override default eph_mode (self.eph_mode).
            Must be '2body' or 'nbody'.

        Returns
        -------
        ephs : `pd.Dataframe`
            Results from propigating the orbit(s) to the specified times.
            Columns like:
            obj_id, sedname, time, ra, dec, dradt, ddecdt, phase, solarelon.
        """
        if not hasattr(self, "ephems"):
            self.setup_ephemerides()
        if eph_mode is None:
            eph_mode = self.eph_mode
        if eph_type is None:
            eph_type = self.eph_type
        self.ephems.set_orbits(sso)
        ephs = self.ephems.generate_ephemerides(
            times,
            time_scale=self.obs_time_scale,
            obscode=self.obs_code,
            eph_mode=eph_mode,
            eph_type=eph_type,
            by_object=True,
        )
        return ephs

    def calc_trailing_losses(self, velocity, seeing, texp=30.0):
        """Calculate the detection and SNR trailing losses.

        'Trailing' losses = loss in sensitivity due to the photons from the
        source being spread over more pixels; thus more sky background is
        included when calculating the flux from the object and thus the SNR
        is lower than for an equivalent brightness stationary/PSF-like source.
        dmagTrail represents this loss.

        'Detection' trailing losses = loss in sensitivity due to the photons
        from the source being spread over more pixels, in a non-stellar-PSF
        way, while source detection is (typically) done using a stellar PSF
        filter and 5-sigma cutoff values based on assuming peaks from
        stellar PSF's above the background; thus the SNR is lower than for an
        equivalent brightness stationary/PSF-like source (and by a greater
        factor than just the simple SNR trailing loss above).
        dmag_detect represents this loss.

        Parameters
        ----------
        velocity : `np.ndarray` or `float`
            The velocity of the moving objects, in deg/day.
        seeing : `np.ndarray` or `float`
            The seeing of the images, in arcseconds.
        texp : `np.ndarray` or `float`, optional
            The exposure time of the images, in seconds. Default 30.

        Returns
        -------
        dmag Trail, dmag_detect : (`np.ndarray` `np.ndarray`)
        or (`float`, `float`)
            dmag_trail and dmag_detect for each set of
            velocity/seeing/texp values.
        """
        a_trail = 0.761
        b_trail = 1.162
        a_det = 0.420
        b_det = 0.003
        x = velocity * texp / seeing / 24.0
        dmag_trail = 1.25 * np.log10(1 + a_trail * x**2 / (1 + b_trail * x))
        dmag_detect = 1.25 * np.log10(1 + a_det * x**2 / (1 + b_det * x))
        return (dmag_trail, dmag_detect)

    def read_filters(
        self,
        filter_dir=None,
        bandpass_root="total_",
        bandpass_suffix=".dat",
        filterlist=("u", "g", "r", "i", "z", "y"),
        v_dir=None,
        v_filter="harris_V.dat",
    ):
        """Read (LSST) and Harris (V) filter throughput curves.

        Only the defaults are LSST specific;
        this can easily be adapted for any survey.

        Parameters
        ----------
        filter_dir : `str`, optional
            Directory containing the filter throughput curves ('total*.dat')
            Default set by 'LSST_THROUGHPUTS_BASELINE' env variable.
        bandpass_root : `str`, optional
            Rootname of the throughput curves in filterlist.
            E.g. throughput curve names are bandpass_root + filterlist[i]
            + bandpass_suffix
            Default `total_` (appropriate for LSST throughput repo).
        bandpass_suffix : `str`, optional
            Suffix for the throughput curves in filterlist.
            Default '.dat' (appropriate for LSST throughput repo).
        filterlist : `list`, optional
            List containing the filter names to use to calculate colors.
            Default ('u', 'g', 'r', 'i', 'z', 'y')
        v_dir : `str`, optional
            Directory containing the V band throughput curve.
            Default None = $RUBIN_SIM_DATA_DIR/movingObjects
        v_filter : `str`, optional
            Name of the V band filter curve.
            Default harris_V.dat.
        """
        if filter_dir is None:
            filter_dir = os.path.join(get_data_dir(), "throughputs/baseline")
        if v_dir is None:
            v_dir = os.path.join(get_data_dir(), "movingObjects")
        self.filterlist = filterlist
        # Read filter throughput curves from disk.
        self.lsst = {}
        for f in self.filterlist:
            self.lsst[f] = Bandpass()
            self.lsst[f].read_throughput(os.path.join(filter_dir, bandpass_root + f + bandpass_suffix))
        self.vband = Bandpass()
        self.vband.read_throughput(os.path.join(v_dir, v_filter))

    def calc_colors(self, sedname="C.dat", sed_dir=None):
        """Calculate the colors for a given SED.

        If the sedname is not already in the dictionary self.colors,
        this reads the SED from disk and calculates all V-[filter] colors
        for all filters in self.filterlist.
        The result is stored in self.colors[sedname][filter], so will not
        be recalculated if the SED + color is reused for another object.

        Parameters
        ----------
        sedname : `str`, optional
            Name of the SED. Default 'C.dat'.
        sed_dir : `str`, optional
            Directory containing the SEDs of the moving objects.
            Default None = $RUBIN_SIM_DATA_DIR/movingObjects,

        Returns
        -------
        colors : `dict` {'filter': color}}
            Dictionary of the colors in self.filterlist.
        """
        if sedname not in self.colors:
            if sed_dir is None:
                sed_dir = os.path.join(get_data_dir(), "movingObjects")
            mo_sed = Sed()
            mo_sed.read_sed_flambda(os.path.join(sed_dir, sedname))
            vmag = mo_sed.calc_mag(self.vband)
            self.colors[sedname] = {}
            for f in self.filterlist:
                self.colors[sedname][f] = mo_sed.calc_mag(self.lsst[f]) - vmag
        return self.colors[sedname]

    def sso_in_circle_fov(self, ephems, obs_data):
        """Determine which observations are within a circular fov
        for a series of observations.
        Note that ephems and obs_data must be the same length.

        Parameters
        ----------
        ephems : `np.recarray`
            Ephemerides for the objects.
        obs_data : `np.recarray`
            The observation pointings.

        Returns
        -------
        indices : `np.ndarray`
            Returns the indexes of the numpy array of the object
            observations which are inside the fov.
        """
        return self._sso_in_circle_fov(ephems, obs_data, self.r_fov)

    def _sso_in_circle_fov(self, ephems, obs_data, r_fov):
        if not self.obs_degrees:
            sep = angular_separation(
                ephems["ra"],
                ephems["dec"],
                np.degrees(obs_data[self.obs_ra]),
                np.degrees(obs_data[self.obs_dec]),
            )
        else:
            sep = angular_separation(
                ephems["ra"],
                ephems["dec"],
                obs_data[self.obs_ra],
                obs_data[self.obs_dec],
            )
        idx_obs = np.where(sep <= r_fov)[0]
        return idx_obs

    def sso_in_rectangle_fov(self, ephems, obs_data):
        """Determine which observations are within a rectangular FoV
        for a series of observations.
        Note that ephems and obs_data must be the same length.

        Parameters
        ----------
        ephems : `np.recarray`
            Ephemerides for the objects.
        obs_data : `np.recarray`
            The observation pointings.

        Returns
        -------
        indices : `np.ndarray`
            Returns the indexes of the numpy array of the object
            observations which are inside the fov.
        """
        return self._sso_in_rectangle_fov(ephems, obs_data, self.x_tol, self.y_tol)

    def _sso_in_rectangle_fov(self, ephems, obs_data, x_tol, y_tol):
        delta_dec = np.abs(ephems["dec"] - obs_data[self.obs_dec])
        delta_ra = np.abs((ephems["ra"] - obs_data[self.obs_ra]) * np.cos(np.radians(obs_data[self.obs_dec])))
        idx_obs = np.where((delta_dec <= y_tol) & (delta_ra <= x_tol))[0]
        return idx_obs

    def sso_in_camera_fov(self, ephems, obs_data):
        """Determine which observations are within the actual
        camera footprint for a series of observations.
        Note that ephems and obs_data must be the same length.

        Parameters
        ----------
        ephems : `np.ndarray`
            Ephemerides for the objects.
        obs_data : `np.ndarray`
            Observation pointings.

        Returns
        -------
        indices : `np.ndarray`
            Returns the indexes of the numpy array of the object
            observations which are inside the fov.
        """
        if not hasattr(self, "camera"):
            self._setup_camera()

        if not self.obs_degrees:
            idx = self.camera(
                ephems["ra"],
                ephems["dec"],
                np.degrees(obs_data[self.obs_ra]),
                np.degrees(obs_data[self.obs_dec]),
                np.degrees(obs_data[self.obs_rot_sky_pos]),
            )
        else:
            idx = self.camera(
                ephems["ra"],
                ephems["dec"],
                obs_data[self.obs_ra],
                obs_data[self.obs_dec],
                obs_data[self.obs_rot_sky_pos],
            )
        return idx

    def sso_in_fov(self, ephems, obs_data):
        """Convenience layer - determine which footprint method to
        apply (from self.footprint) and use it.

        Parameters
        ----------
        ephems : `np.ndarray`
            Ephemerides for the objects.
        obs_data : `np.ndarray`
            Observation pointings.

        Returns
        -------
        indices : `np.ndarray`
            Returns the indexes of the numpy array of the object
            observations which are inside the fov.
        """
        if self.footprint == "camera":
            return self.sso_in_camera_fov(ephems, obs_data)
        elif self.footprint == "rectangle":
            return self.sso_in_rectangle_fov(ephems, obs_data)
        elif self.footprint == "circle":
            return self.sso_in_circle_fov(ephems, obs_data)
        else:
            warnings.warn("Using circular fov; could not match specified footprint.")
            self.footprint = "circle"
            return self.sso_in_circle_fov(ephems, obs_data)



__all__ = (
    "generate_microlensing_slicer",
    "MicrolensingMetric",
    "microlensing_amplification",
    "microlensing_amplification_fsfb",
    "info_peak_before_t0",
    "fisher_matrix",
    "coefficients_pspl",
    "coefficients_fsfb",
)

import os
from copy import deepcopy

import numpy as np
from astropy import units as u
from astropy.coordinates import SkyCoord
from rubin_scheduler.data import get_data_dir
from rubin_scheduler.utils import hpid2_ra_dec

import rubin_sim.maf.metrics as metrics
import rubin_sim.maf.slicers as slicers
from rubin_sim.maf.utils import m52snr


# Via Natasha Abrams nsabrams@college.harvard.edu
def microlensing_amplification(t, impact_parameter=1, crossing_time=1825.0, peak_time=100, blending_factor=1):
    """The microlensing amplification

    Parameters
    ----------
    t : `float`
        The time of observation (days)
    impact_parameter : `float`
        The impact paramter (0 means big amplification)
    crossing_time : `float`
        Einstein crossing time (days)
    peak_time : `float`
        The peak time (days)
    blending_factor: `float`
        The blending factor where 1 is unblended
    """

    lightcurve_u = np.sqrt(impact_parameter**2 + ((t - peak_time) ** 2 / crossing_time**2))
    amplified_mag = (lightcurve_u**2 + 2) / (
        lightcurve_u * np.sqrt(lightcurve_u**2 + 4)
    ) * blending_factor + (1 - blending_factor)

    return amplified_mag


def microlensing_amplification_fsfb(t, impact_parameter=1, crossing_time=1825.0, peak_time=100, fs=20, fb=20):
    """The microlensing amplification in terms of source flux and blend flux

    Parameters
    ----------
    t : `float`
        The time of observation (days)
    impact_parameter : `float`
        The impact paramter (0 means big amplification)
    crossing_time : `float`
        Einstein crossing time (days)
    peak_time : `float`
        The peak time (days)
    """

    lightcurve_u = np.sqrt(impact_parameter**2 + ((t - peak_time) ** 2 / crossing_time**2))
    amplified_mag = (lightcurve_u**2 + 2) / (lightcurve_u * np.sqrt(lightcurve_u**2 + 4)) * fs + fb

    return amplified_mag


def info_peak_before_t0(impact_parameter=1, crossing_time=100.0):
    """Time of Maximum Information before peak

    The point of maximum information before peak indicates when in a
    lightcurve with constant snr the contribution to the
    characterization of the event timescale tE is at maximum. After that
    point in timem the timescale characterization is decisively improved.
    via Markus Hundertmark markus.hundertmark@uni-heidelberg.de

    Parameters
    ----------
    impact_parameter : `float`
        The impact paramter (0 means big amplification)
    crossing_time : `float`
        Einstein crossing time (days)
    """

    optimal_time = (
        crossing_time
        * np.sqrt(
            -(impact_parameter**2) + np.sqrt(9 * impact_parameter**4 + 36 * impact_parameter**2 + 4) - 2
        )
        / 2
    )
    return np.array(optimal_time)


def coefficients_pspl(t, t0, te, u0, fs, fb):
    """Coefficients for calculating the first derivatives wrt t0,te,u0
    in the Fisher matrix in an effective way, optimized with sympy and cse
    via Markus Hundertmark markus.hundertmark@uni-heidelberg.de

    Parameters
    ----------
    t : `float`
        The time of observation (days usually as JD or HJD)
    u0 : `float`
        The impact parameter (0 means high magnification)
    te : `float`
        Einstein crossing time (days)
    t0 : `float`
        Time of peak (days usually as JD or HJD)
    fs : `float`
        Source flux (counts/s but here fs = 10.**(-0.4*mag_source)
        in the respective passband)
    fb : `float`
        Blend flux (counts/s but here fs = 10.**(-0.4*mag_blend)
        in the respective passband)
    """
    x0 = te ** (-2)
    x1 = (t - t0) ** 2
    x2 = u0**2 + x0 * x1
    x3 = 1 / np.sqrt(x2)
    x4 = 2 * x3
    x5 = x2 + 4
    x6 = fs / np.sqrt(x5)
    x7 = x1 / te**3
    x8 = x6 * x7
    x9 = x2 + 2
    x10 = x9 / x2 ** (3 / 2)
    x11 = fs * x3 * x9 / x5 ** (3 / 2)
    x12 = u0 * x6
    x13 = x0 * (-2 * t + 2 * t0)
    x14 = x13 * x6

    c0 = x10 * x8 + x11 * x7 - x4 * x8
    c1 = -u0 * x11 - x10 * x12 + x12 * x4
    c2 = -1 / 2 * x10 * x14 - 1 / 2 * x11 * x13 + x14 * x3

    return [c2, c1, c0]


def coefficients_fsfb(t, t0, te, u0, fs, fb):
    """Coefficients for calculating the first derivatives wrt fs,fb
    in the Fisher (information) matrix in an effective way, optimized
    with sympy and cse this function needs to be evaluated for each
    passband respectively.
    via Markus Hundertmark markus.hundertmark@uni-heidelberg.de

    Parameters
    ----------
    t : `float`
        The time of observation (days usually as JD or HJD)
    u0 : `float`
        The impact parameter (0 means high magnification)
    te : `float`
        Einstein crossing time (days)
    t0 : `float`
        Time of peak (days usually as JD or HJD)
    fs : `float`
        Source flux (counts/s but here fs = 10.**(-0.4*mag_source)
        in the respective passband)
    fb : `float`
        Blend flux (counts/s but here fs = 10.**(-0.4*mag_blend)
        in the respective passband)
    """
    x0 = u0**2 + (t - t0) ** 2 / te**2
    c0 = (x0 + 2) / (np.sqrt(x0) * np.sqrt(x0 + 4))
    c1 = 1
    return [c1, c0]


def fisher_matrix(t, t0, te, u0, fs, fb, snr, filters="ugriz", filter_name="i"):
    """The Fisher (information) matrix relying on first derivatives
    wrt t0,te,u0,fs,fb, relying on with cse optimized coefficents.
    This implementation for the non-linear magnification is subject to the
    respective underlying assumptions, i.e. relatively small uncertainties.
    NB the Fisher matrix is using first derivatives, alternatively one
    could consider second derivatives the resulting covariance matrix as
    inverse of the Fisher matrix is a Cramer Rao lower bound of the
    respective uncertainty.
    The function returns the full information matrix for all t.

    via Markus Hundertmark markus.hundertmark@uni-heidelberg.de

    Parameters
    ----------
    t : `float`
        The time of observation (days usually as JD or HJD)
    u0 : `float`
        The impact parameter (0 means high magnification)
    te : `float`
        Einstein crossing time (days)
    t0 : `float`
        Time of peak (days usually as JD or HJD)
    fs : `float`
        Source flux (counts/s but here fs = 10.**(-0.4*mag_source)
        in the respective passband)
    fb : `float`
        Blend flux (counts/s but here fs = 10.**(-0.4*mag_blend)
        in the respective passband)
    snr : `float`
        Signal to noise ratio in order to infer the flux uncertainty
    """
    # indices are 0:te, 1:u0, 2:t0
    # 3:fs_i, 4:fb_i, 5:fs_i+1, 6: fb_i+1, etc. for ugrizy
    filters = filters
    npars = 2 * len(filters) + 3
    fis_mat = np.zeros((npars, npars))
    filter_index = np.where(filters == filter_name)[0]
    # initialize result matrix
    for i in range(len(t)):
        t_value = t[i]
        # err = flux_err[i]
        matrix = np.zeros((npars, npars))
        derivatives = coefficients_pspl(t_value, t0, te, u0, fs, fb)
        fsfb_derivatives = 2 * len(filters) * [0]
        fbdiff, fsdiff = coefficients_fsfb(t_value, t0, te, u0, fs, fb)
        np.array(fsfb_derivatives)[filter_index * 2] = fsdiff
        np.array(fsfb_derivatives)[filter_index * 2 + 1] = fbdiff
        derivatives = derivatives + fsfb_derivatives
        for idx in range(npars):
            for jdx in range(idx, npars):
                matrix[idx, jdx] = derivatives[idx] * derivatives[jdx]
        # restore symmetrical entries
        matrix = matrix + matrix.T - np.diag(np.diag(matrix))

        # microlenisng_amplification function to replace
        usqr = (t_value - t0) ** 2 / te**2 + u0**2
        mu = (usqr + 2.0) / (usqr * (usqr + 4)) ** 0.5
        # may need to check shape of snr and t
        flux_err = (fs * mu + fb) / snr[i]

        matrix = matrix / (flux_err**2)
        fis_mat = fis_mat + matrix
    return fis_mat


class MicrolensingMetric(metrics.BaseMetric):
    """
    Quantifies detectability of Microlensing events.
    Can also return the number of datapoints within two crossing times
    of the peak of event.

    Parameters
    ----------
    metric_calc : `str`
        Type of metric.
        If metric_calc == 'detect', returns the number of microlensing events
        detected within certain parameters.
        If metric_calc == 'Npts', returns the number of points
        within two crossing times of the peak of the vent.
    pts_needed : `int`
        Number of an object's lightcurve points required to be above the
        5-sigma limiting depth before it is considered detected.
    time_before_peak: `int` or `str`
        Number of days before lightcurve peak to qualify event as triggered.
        If time_before_peak == 'optimal', the number of days before the
        lightcurve peak is the time of maximal information.
    detect: `bool`
        By default we trigger which only looks at points before the peak
        of the lightcurve. When detect = True, observations on either side
        of the lightcurve are considered.

    Notes
    -----
    Expects slice_point to have keys of:
        peak_time : float (days)
        crossing_time : float (days)
        impact_parameter : float (positive)
        blending_factors (optional): float (between 0 and 1)
    """

    def __init__(
        self,
        metric_name="MicrolensingMetric",
        metric_calc="detect",
        mjd_col="observationStartMJD",
        m5_col="fiveSigmaDepth",
        filter_col="filter",
        night_col="night",
        pts_needed=2,
        mag_dict=None,
        detect_sigma=3.0,
        time_before_peak=0,
        detect=False,
        **kwargs,
    ):
        self.metric_calc = metric_calc
        if metric_calc == "detect":
            self.units = "Detected, 0 or 1"
        elif metric_calc == "Npts":
            self.units = "Npts within 2tE"
        elif metric_calc == "Fisher":
            self.units = "Characterized, 0 or 1"
        else:
            raise Exception('metric_calc must be "detect", "Npts" or "Fisher".')
        if mag_dict is None:
            # Mean mag of stars in each filter from TRIstarDensity map
            mag_dict = {
                "u": 25.2,
                "g": 25.0,
                "r": 24.5,
                "i": 23.4,
                "z": 22.8,
                "y": 22.5,
            }
        self.mjd_col = mjd_col
        self.m5_col = m5_col
        self.filter_col = filter_col
        self.night_col = night_col
        self.pts_needed = pts_needed
        self.detect_sigma = detect_sigma
        self.time_before_peak = time_before_peak
        self.detect = detect
        self.mags = mag_dict

        cols = [self.mjd_col, self.m5_col, self.filter_col, self.night_col]
        super(MicrolensingMetric, self).__init__(
            col=cols, units=self.units, metric_name=metric_name + "_" + self.metric_calc, **kwargs
        )

    def run(self, data_slice, slice_point=None):
        if self.detect is True and self.time_before_peak > 0:
            raise Exception("When detect = True, time_before_peak must be zero")

        # Generate the lightcurve for this object
        # make t a kind of simple way
        t = data_slice[self.mjd_col] - np.min(data_slice[self.night_col])
        t = t - t.min()

        #
        # Try for if a blending factor slice was added
        if "apparent_m_no_blend_u" in slice_point:
            amplitudes = np.zeros(len(t))
            individual_mags = True
        else:
            if "blending_factor" in slice_point:
                amplitudes = microlensing_amplification(
                    t,
                    impact_parameter=slice_point["impact_parameter"],
                    crossing_time=slice_point["crossing_time"],
                    peak_time=slice_point["peak_time"],
                    blending_factor=slice_point["blending_factor"],
                )
                individual_mags = False
            else:
                amplitudes = microlensing_amplification(
                    t,
                    impact_parameter=slice_point["impact_parameter"],
                    crossing_time=slice_point["crossing_time"],
                    peak_time=slice_point["peak_time"],
                )
                individual_mags = False

        filters = np.unique(data_slice[self.filter_col])
        amplified_mags = amplitudes * 0

        npars = 2 * len(filters) + 3
        fis_mat = np.zeros((npars, npars))
        fis_mat_wzeros = deepcopy(fis_mat)

        for filtername in filters:
            infilt = np.where(data_slice[self.filter_col] == filtername)[0]

            if individual_mags is True:
                fs = slice_point["apparent_m_no_blend_{}".format(filtername)]
                fb = slice_point["apparent_m_{}".format(filtername)] - fs
                amplitudes = microlensing_amplification_fsfb(
                    t,
                    impact_parameter=slice_point["impact_parameter"],
                    crossing_time=slice_point["crossing_time"],
                    peak_time=slice_point["peak_time"],
                    fs=fs,
                    fb=fb,
                )
                self.mags[filtername] = slice_point["apparent_m_{}".format(filtername)]
                amplified_mags[infilt] = self.mags[filtername] - 2.5 * np.log10(amplitudes[infilt])

            else:
                amplified_mags[infilt] = self.mags[filtername] - 2.5 * np.log10(amplitudes[infilt])

        # The SNR of each point in the light curve
        snr = m52snr(amplified_mags, data_slice[self.m5_col])
        # The magnitude uncertainties that go with amplified mags
        mag_uncert = 2.5 * np.log10(1 + 1.0 / snr)

        n_pre = []
        n_post = []
        for filtername in filters:
            if self.metric_calc == "detect":
                if self.time_before_peak == "optimal":
                    time_before_peak_optimal = info_peak_before_t0(
                        slice_point["impact_parameter"], slice_point["crossing_time"]
                    )
                    # observations pre-peak and in the given filter
                    infilt = np.where(
                        (data_slice[self.filter_col] == filtername)
                        & (t < (slice_point["peak_time"] - time_before_peak_optimal))
                    )[0]

                else:
                    # observations pre-peak and in the given filter
                    infilt = np.where(
                        (data_slice[self.filter_col] == filtername)
                        & (t < (slice_point["peak_time"] - self.time_before_peak))
                    )[0]

                # observations post-peak and in the given filter
                outfilt = np.where(
                    (data_slice[self.filter_col] == filtername) & (t > slice_point["peak_time"])
                )[0]
                # Broadcast to calc the mag_i - mag_j
                diffs = amplified_mags[infilt] - amplified_mags[infilt][:, np.newaxis]
                diffs_uncert = np.sqrt(mag_uncert[infilt] ** 2 + mag_uncert[infilt][:, np.newaxis] ** 2)
                diffs_post = amplified_mags[outfilt] - amplified_mags[outfilt][:, np.newaxis]
                diffs_post_uncert = np.sqrt(
                    mag_uncert[outfilt] ** 2 + mag_uncert[outfilt][:, np.newaxis] ** 2
                )

                # Calculating this as a catalog-level detection. In theory,
                # we could have a high SNR template image, so there would be
                # little to no additional uncertianty from the subtraction.

                sigma_above = np.abs(diffs) / diffs_uncert
                sigma_above_post = np.abs(diffs_post) / diffs_post_uncert
                # divide by 2 because array has i,j and j,i
                n_above = np.size(np.where(sigma_above > self.detect_sigma)[0]) / 2
                n_pre.append(n_above)
                n_above_post = np.size(np.where(sigma_above_post > self.detect_sigma)[0]) / 2
                n_post.append(n_above_post)

            elif self.metric_calc == "Npts":
                # observations pre-peak and in the given filter within 2tE
                infilt = np.where(
                    (data_slice[self.filter_col] == filtername)
                    & (t < (slice_point["peak_time"]))
                    & (t > (slice_point["peak_time"] - slice_point["crossing_time"]))
                    & (snr > self.detect_sigma)
                )[0]
                # observations post-peak and in the given filter within 2tE
                outfilt = np.where(
                    (data_slice[self.filter_col] == filtername)
                    & (t > (slice_point["peak_time"]))
                    & (t < (slice_point["peak_time"] + slice_point["crossing_time"]))
                    & (snr > self.detect_sigma)
                )[0]
                n_pre.append(len(infilt))
                n_post.append(len(outfilt))

            elif self.metric_calc == "Fisher":
                if individual_mags is True:
                    fs = slice_point["apparent_m_no_blend_{}".format(filtername)]
                    fb = slice_point["apparent_m_{}".format(filtername)] - fs
                else:
                    fs = self.mags[filtername]
                    # assuming that in populated areas of the galaxy
                    # the blend fraction is 0.5
                    fb = fs

                fis_mat = fis_mat_wzeros + fisher_matrix(
                    t[infilt],
                    t0=slice_point["peak_time"],
                    te=slice_point["crossing_time"],
                    u0=slice_point["impact_parameter"],
                    fs=fs,
                    fb=fb,
                    snr=snr[infilt],
                    filters=filters,
                    filter_name=filtername,
                )

                # We remove entries if there's no data in a filter
                # which would keep it from being inverted
                # detectable information in dimensions:
                fis_mat_wzeros = deepcopy(fis_mat)
                mask_idx = np.where(np.diag(fis_mat) == 0)
                fis_mat = np.delete(fis_mat, mask_idx[0], 0)
                fis_mat = np.delete(fis_mat, mask_idx[0], 1)
                try:
                    cov = np.linalg.inv(fis_mat)
                    sigmat_e_t_e = cov[0, 0] ** 0.5 / slice_point["crossing_time"]
                    # sigmau0_u0 = cov[1, 1] ** 0.5
                    #             / slice_point["impact_parameter"]
                    # corr_btwn_t_eu0 = cov[0, 1] /
                    #    (slice_point["crossing_time"]
                    #    * slice_point["impact_parameter"])
                except np.linalg.LinAlgError:
                    sigmat_e_t_e = np.inf
                    # sigmau0_u0 = np.inf
                    # corr_btwn_t_eu0 = np.inf

        npts = np.sum(n_pre)
        npts_post = np.sum(n_post)
        if self.metric_calc == "detect":
            if self.detect is True:
                if npts >= self.pts_needed and npts_post >= self.pts_needed:
                    return 1
                else:
                    return 0
            else:
                if npts >= self.pts_needed:
                    return 1
                else:
                    return 0
        elif self.metric_calc == "Npts":
            return npts + npts_post
        elif self.metric_calc == "Fisher":
            # cutoff of sigmat_e_t_e < 0.1 for
            # values determined by the 3sigma of
            # pubished planet candidates
            return sigmat_e_t_e


def generate_microlensing_slicer(
    min_crossing_time=1,
    max_crossing_time=10,
    t_start=1,
    t_end=3652,
    n_events=10000,
    seed=42,
    nside=128,
    filtername="r",
):
    """Generate a UserPointSlicer with a population of microlensing events.
    To be used with MicrolensingMetric

    Parameters
    ----------
    min_crossing_time : `float`
        The minimum crossing time for the events generated (days)
    max_crossing_time : `float`
        The max crossing time for the events generated (days)
    t_start : `float`
        The night to start generating peaks (days)
    t_end : `float`
        The night to end generating peaks (days)
    n_events : `int`
        Number of microlensing events to generate
    seed : `float`
        Random number seed
    nside : `int`
        HEALpix nside, used to pick which stellar density map to load
    filtername : `str`
        The filter to use for the stellar density map

    Returns
    -------
    slicer : `maf.UserPointsSlicer`
        A slicer populated by microlensing events
        (each slice_point is a different event)
    """
    # Seed the random number generator and generate random parameters
    rng = np.random.default_rng(seed)
    crossing_times = rng.uniform(low=min_crossing_time, high=max_crossing_time, size=n_events)
    peak_times = rng.uniform(low=t_start, high=t_end, size=n_events)
    impact_paramters = rng.uniform(low=0, high=1, size=n_events)

    map_dir = os.path.join(get_data_dir(), "maps", "TriMaps")
    data = np.load(os.path.join(map_dir, "TRIstarDensity_%s_nside_%i.npz" % (filtername, nside)))
    star_density = data["starDensity"].copy()
    # magnitude bins
    bins = data["bins"].copy()
    data.close()

    star_mag = 22
    bin_indx = np.where(bins[1:] >= star_mag)[0].min()
    density_used = star_density[:, bin_indx].ravel()
    order = np.argsort(density_used)
    # I think the model might have a few outliers at the extreme,
    # let's truncate it a bit
    density_used[order[-10:]] = density_used[order[-11]]

    # now, let's draw N from that distribution squared
    dist = density_used[order] ** 2
    cumm_dist = np.cumsum(dist)
    cumm_dist = cumm_dist / np.max(cumm_dist)
    uniform_draw = rng.uniform(size=n_events)
    indexes = np.floor(np.interp(uniform_draw, cumm_dist, np.arange(cumm_dist.size)))
    hp_ids = order[indexes.astype(int)]
    gal_l, gal_b = hpid2_ra_dec(nside, hp_ids, nest=True)
    c = SkyCoord(l=gal_l * u.deg, b=gal_b * u.deg, frame="galactic").transform_to("icrs")
    ra, dec = c.ra.deg, c.dec.deg
    # Set up the slicer to evaluate the catalog we just made
    slicer = slicers.UserPointsSlicer(ra, dec, lat_lon_deg=True, badval=0)
    # Add any additional information about each object to the slicer
    slicer.slice_points["peak_time"] = peak_times
    slicer.slice_points["crossing_time"] = crossing_times
    slicer.slice_points["impact_parameter"] = impact_paramters

    return slicer



#
# LSST Data Management System
# Copyright 2008, 2009, 2010 LSST Corporation.
#
# This product includes software developed by the
# LSST Project (http://www.lsst.org/).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the LSST License Statement and
# the GNU General Public License along with this program.  If not,
# see <http://www.lsstcorp.org/LegalNotices/>.
#
"""Python interface to the metrics analysis framework.
"""
from .batches import *
from .db import *
from .maf_contrib import *
from .maps import *
from .metric_bundles import *
from .metrics import *
from .plots import *
from .run_comparison import *
from .slicers import *
from .stackers import *
from .utils import *
from .web import *



import unittest

import numpy as np

import rubin_sim.maf.metrics as metrics


class TestTemplateMetrics(unittest.TestCase):
    def test_incremental_template_metric(self):
        """
        Test the incremental template metric
        """
        npoints = 10
        names = ["seeingFwhmEff", "fiveSigmaDepth", "observationStartMJD", "night", "filter"]
        types = [float, float, float, int, str]
        data = np.zeros(npoints, dtype=list(zip(names, types)))
        data["seeingFwhmEff"] = [0.7, 0.5, 1.5, 1.2, 0.6, 0.65, 1.2, 1.2, 1.2, 1.2]
        data["fiveSigmaDepth"] = [24.5, 24.5, 23, 23.5, 23.5, 24.5, 23, 23, 23, 23]
        data["observationStartMJD"] = np.arange(npoints)
        data["night"] = np.arange(npoints)
        data["filter"] = ["r" for i in range(npoints)]

        ttm = metrics.TemplateTime(n_images_in_template=3, seeing_percentile=50, m5_percentile=50)
        metric_val = ttm.run(data)

        night_created = ttm.reduce_Night_template_created(metric_val)
        n_nights_no_template = ttm.reduce_N_nights_without_template(metric_val)
        n_images_until_template = ttm.reduce_N_images_until_template(metric_val)
        n_images_with_template = ttm.reduce_N_images_with_template(metric_val)
        template_m5 = ttm.reduce_Template_m5(metric_val)
        template_seeing = ttm.reduce_Template_seeing(metric_val)

        self.assertEqual(night_created, 4)
        self.assertEqual(n_nights_no_template, 4)
        self.assertEqual(n_images_until_template, 5)
        self.assertEqual(n_images_with_template, 5)
        self.assertAlmostEqual(template_m5, 24.917687379997627)
        self.assertEqual(template_seeing, 0.607672394977614)

        # alternative values

        ttm = metrics.TemplateTime(n_images_in_template=4, seeing_percentile=75, m5_percentile=75)
        metric_val = ttm.run(data)

        night_created = ttm.reduce_Night_template_created(metric_val)
        n_nights_no_template = ttm.reduce_N_nights_without_template(metric_val)
        n_images_until_template = ttm.reduce_N_images_until_template(metric_val)
        n_images_with_template = ttm.reduce_N_images_with_template(metric_val)
        template_m5 = ttm.reduce_Template_m5(metric_val)
        template_seeing = ttm.reduce_Template_seeing(metric_val)

        self.assertEqual(night_created, 5)
        self.assertEqual(n_nights_no_template, 5)
        self.assertEqual(n_images_until_template, 6)
        self.assertEqual(n_images_with_template, 4)
        self.assertAlmostEqual(template_m5, 25.124349265777216)
        self.assertEqual(template_seeing, 0.6213856145706174)

        # test no template created

        ttm = metrics.TemplateTime(n_images_in_template=3, seeing_percentile=25, m5_percentile=25)
        metric_val = ttm.run(data)

        self.assertEqual(metric_val, -666)


if __name__ == "__main__":
    unittest.main()



__all__ = ("name_sanitize",)


def name_sanitize(in_string):
    """
    Convert a string to a more file name (and web) friendly format.

    Parameters
    ----------
    in_string : `str`
        The input string to be sanitized.
        Typically these are combinations of metric names and metadata.

    Returns
    -------
    out_string : `str`
        The string after removal/replacement of non-friendly characters.
    """
    # Replace <, > and = signs.
    out_string = in_string.replace(">", "gt").replace("<", "lt").replace("=", "eq")
    # Remove single-spaces, strip '.'s and ','s
    out_string = out_string.replace(" ", "_").replace(".", "_").replace(",", "")
    # and remove / and \
    out_string = out_string.replace("/", "_").replace("\\", "_")
    # and remove parentheses
    out_string = out_string.replace("(", "").replace(")", "")
    # Remove ':' and ';"
    out_string = out_string.replace(":", "_").replace(";", "_")
    # Replace '%' and #
    out_string = out_string.replace("%", "_").replace("#", "_")
    # Remove '__'
    while "__" in out_string:
        out_string = out_string.replace("__", "_")
    return out_string



__all__ = ("NDSlicer",)

import itertools
import warnings
from functools import wraps

import numpy as np

from rubin_sim.maf.plots.nd_plotters import OneDSubsetData, TwoDSubsetData

from .base_slicer import BaseSlicer

# nd Slicer slices data on N columns in sim_data


class NDSlicer(BaseSlicer):
    """Nd slicer (N dimensions)

    Parameters
    ==========
    slice_col_list : `list` of `str`
        Names of the data columns for slicing (e.g. 'airmass`, etc)
    bins_list : `int` or `list` of `int` or `np.ndarray`, opt
        Single integer (for same number of slices in each dimension)
        or a list of integers (matching
        slice_col_list) or list of arrays. Default 100, in all dimensions.

    All bins are half-open ([a, b)).
    """

    def __init__(self, slice_col_list=None, bins_list=100, verbose=True):
        super().__init__(verbose=verbose)
        self.slice_col_list = slice_col_list
        self.columns_needed = self.slice_col_list
        if self.slice_col_list is not None:
            self.n_d = len(self.slice_col_list)
        else:
            self.n_d = None
        self.bins_list = bins_list
        if not (isinstance(bins_list, float) or isinstance(bins_list, int)):
            if len(self.bins_list) != self.n_d:
                raise Exception(
                    "bins_list must be same length as slice_col_names, unless it is a single value"
                )
        self.slicer_init = {"slice_col_list": slice_col_list, "bins_list": bins_list}
        self.plot_funcs = [TwoDSubsetData, OneDSubsetData]

    def setup_slicer(self, sim_data, maps=None):
        """Set up bins."""
        # Parse input bins choices.
        self.bins = []
        # If we were given a single number for the binsList, convert to list.
        if isinstance(self.bins_list, float) or isinstance(self.bins_list, int):
            self.bins_list = [self.bins_list for c in self.slice_col_list]
        # And then build bins.
        for bl, col in zip(self.bins_list, self.slice_col_list):
            if isinstance(bl, float) or isinstance(bl, int):
                slice_col = sim_data[col]
                bin_min = slice_col.min()
                bin_max = slice_col.max()
                if bin_min == bin_max:
                    warnings.warn("bin_min=bin_max for column %s: increasing bin_max by 1." % (col))
                    bin_max = bin_max + 1
                bin_size = (bin_max - bin_min) / float(bl)
                bins = np.arange(bin_min, bin_max + bin_size / 2.0, bin_size, "float")
                self.bins.append(bins)
            else:
                self.bins.append(np.sort(bl))
        self.nslice = (np.array(list(map(len, self.bins))) - 1).prod()
        # Count how many bins we have total
        # (not counting last 'RHS' bin values, as in oneDSlicer).
        self.shape = self.nslice
        # Set up slice metadata.
        self.slice_points["sid"] = np.arange(self.nslice)
        # Including multi-D 'leftmost' bin values
        bins_for_iteration = []
        for b in self.bins:
            bins_for_iteration.append(b[:-1])
        biniterator = itertools.product(*bins_for_iteration)
        self.slice_points["bins"] = []
        for b in biniterator:
            self.slice_points["bins"].append(b)
        # and multi-D 'leftmost' bin indexes corresponding to each sid
        self.slice_points["bin_idxs"] = []
        bin_ids_for_iteration = []
        for b in self.bins:
            bin_ids_for_iteration.append(np.arange(len(b[:-1])))
        bin_id_iterator = itertools.product(*bin_ids_for_iteration)
        for bidx in bin_id_iterator:
            self.slice_points["bin_idxs"].append(bidx)
        # Add metadata from maps.
        self._run_maps(maps)
        # Set up indexing for data slicing.
        self.sim_idxs = []
        self.lefts = []
        for slice_col_name, bins in zip(self.slice_col_list, self.bins):
            sim_idxs = np.argsort(sim_data[slice_col_name])
            sim_fields_sorted = np.sort(sim_data[slice_col_name])
            # "left" values are location where simdata == bin value
            left = np.searchsorted(sim_fields_sorted, bins[:-1], "left")
            left = np.concatenate(
                (
                    left,
                    np.array(
                        [
                            len(sim_idxs),
                        ]
                    ),
                )
            )
            # Add these calculated values into the class lists of
            # sim_idxs and lefts.
            self.sim_idxs.append(sim_idxs)
            self.lefts.append(left)

        @wraps(self._slice_sim_data)
        def _slice_sim_data(islice):
            """Slice sim_data to return relevant indexes for slice_point."""
            # Identify relevant pointings in each dimension.
            sim_idxs_list = []
            # Translate islice into indexes in each bin dimension
            bin_idxs = self.slice_points["bin_idxs"][islice]
            for d, i in zip(list(range(self.n_d)), bin_idxs):
                sim_idxs_list.append(set(self.sim_idxs[d][self.lefts[d][i] : self.lefts[d][i + 1]]))
            idxs = list(set.intersection(*sim_idxs_list))
            return {
                "idxs": idxs,
                "slice_point": {
                    "sid": islice,
                    "bin_left": self.slice_points["bins"][islice],
                    "bin_idxs": self.slice_points["bin_idxs"][islice],
                },
            }

        setattr(self, "_slice_sim_data", _slice_sim_data)

    def __eq__(self, other_slicer):
        """Evaluate if grids are equivalent."""
        if isinstance(other_slicer, NDSlicer):
            if other_slicer.n_d != self.n_d:
                return False
            for i in range(self.n_d):
                if not np.array_equal(other_slicer.slice_points["bins"][i], self.slice_points["bins"][i]):
                    return False
            return True
        else:
            return False
